;;; -*- Lisp -*-
;;;
;;; Traditional western music notation (CMN=Common Music Notation)
;;;
;;; default font is the Sonata font -- to use Petrucci-Medium, put :Petrucci-Medium in *features* rather than :Sonata


(pushnew :cmn *features*)

#+KCL (in-package :cmn :use '(:lisp :pcl :loop))
#+KCL (shadow 'system)

#+(and Excl (not cltl2)) 
  (progn
    (require :loop)
    (in-package :cmn :use '(:lisp :pcl :loop))
    (import '(pcl::defconstructor) :cmn))

#+(and Excl cltl2)
  (progn
    (defpackage :cmn (:use :common-lisp :clos) (:shadow "REST") (:import-from "EXCL" "RATIOP"))
    (in-package :cmn)
    #-Allegro-v4.1 (import '(clos::defconstructor) :cmn)
    )


#+Lucid (in-package :cmn :use '(:lisp))	;on Sun, Lucid has CLOS and LOOP built-in
					;I don't know about Lucid's foreign function interface

;;; for CCL package definition, see mac-cmn.lisp
#+CCL (in-package :cmn)
#+CCL (progn (shadow 'CCL:COPY) (shadow 'CCL:CANCEL))

#-cltl2 (shadow 'rest)

#+Petrucci-Medium (defparameter Music-Font "Petrucci-Medium")
#+Sonata          (defparameter Music-Font "Sonata")
#-(or Sonata Petrucci-Medium) (print "you forgot to put the font name on the *features* list")



;;; since PCL's make instance is about 10 times slower than it needs to be for our use of it,
;;; most of the make-instance calls are packaged up in functions with names like make-<class-name>
;;; and the actual code called is in cmn4.lisp at the very end of the file -- we need to compile
;;; and load the class definitions before calling PCL's defconstructor.  I wasn't able to get
;;; the defconstructors to compile in kcl -- no room.
;;;
;;; Similarly, it speeds up slot access by about an order of magnitude if the slot has a unique name.
;;; Most functions with "%" in their name are internal optimizations of some sort -- these redundant
;;; names make our code ugly and hard to read, but it's the user's view that matters, I suppose.


#-Excl (defun ratiop (n) (and (not (integerp n)) (typep n 'ratio)))
#+(and excl (not cltl2)) (import '(excl:ratiop))

#+(and Excl NeXT) (progn
	 
	 #-Allegro-v4.1 (load (merge-pathnames "nxtcmn.o" excl::*source-pathname*)
			      #-NeXT-3.0 :system-libraries #-NeXT-3.0 '("sys_s")
			      )
;;; you may have to fix this pathname to point to your version of nxtcmn.o
	 
	 (ff:defforeign 'c-open :entry-point "_OpenCMNFile" :arg-checking nil :arguments '(string) :return-type :integer)
	 (ff:defforeign 'c-close :entry-point "_CloseCMNFile" :arg-checking nil :arguments nil)
	 (ff:defforeign 'c-print :entry-point "_PrintCMNFile" :arg-checking nil :arguments '(string))
	 (ff:defforeign 'c-lineto :entry-point "_LinetoCMNFile" :arg-checking nil :arguments '(array))
	 (ff:defforeign 'c-rlineto :entry-point "_RLinetoCMNFile" :arg-checking nil :arguments '(array))
	 (ff:defforeign 'c-moveto :entry-point "_MovetoCMNFile" :arg-checking nil :arguments '(array))
	 (ff:defforeign 'c-rmoveto :entry-point "_RMovetoCMNFile" :arg-checking nil :arguments '(array))
	 (ff:defforeign 'c-curveto :entry-point "_CurvetoCMNFile" :arg-checking nil :arguments '(array))
	 (ff:defforeign 'c-fill-in :entry-point "_FillinCMNFile" :arg-checking nil :arguments '(array string))
	 (ff:defforeign 'c-glyph :entry-point "_GlyphCMNFile" :arg-checking nil :arguments '(fixnum))
	 (ff:defforeign 'c-font-glyph :entry-point "_FontGlyphCMNFile" :arg-checking nil :arguments '(string fixnum fixnum))
	 (ff:defforeign 'c-text :entry-point "_FontTextCMNFile" :arg-checking nil :arguments '(string fixnum string))
	 (ff:defforeign 'c-stem :entry-point "_StemCMNFile" :arg-checking nil :arguments '(array))
	 (ff:defforeign 'c-just-stem :entry-point "_VerticalLineCMNFile" :arg-checking nil :arguments '(array))
	 (ff:defforeign 'c-line :entry-point "_HorizontalLineCMNFile" :arg-checking nil :arguments '(array))
	 (ff:defforeign 'c-slanted-line :entry-point "_SlantedLineCMNFile" :arg-checking nil :arguments '(array))
	 )

#+(and Excl SGI)
 (progn
 	(load (merge-pathnames "nxtcmn.o" excl::*source-pathname*) )
 	(ff:defforeign 'c-open :entry-point "OpenCMNFile" :arg-checking nil :arguments '(string) :return-type :integer)
 	(ff:defforeign 'c-close :entry-point "CloseCMNFile" :arg-checking nil :arguments nil)
 	(ff:defforeign 'c-print :entry-point "PrintCMNFile" :arg-checking nil :arguments '(string))
 	(ff:defforeign 'c-lineto :entry-point "LinetoCMNFile" :arg-checking nil :arguments '(array))
 	(ff:defforeign 'c-rlineto :entry-point "RLinetoCMNFile" :arg-checking nil :arguments '(array))
 	(ff:defforeign 'c-moveto :entry-point "MovetoCMNFile" :arg-checking nil :arguments '(array))
 	(ff:defforeign 'c-rmoveto :entry-point "RMovetoCMNFile" :arg-checking nil :arguments '(array))
 	(ff:defforeign 'c-curveto :entry-point "CurvetoCMNFile" :arg-checking nil :arguments '(array))
 	(ff:defforeign 'c-fill-in :entry-point "FillinCMNFile" :arg-checking nil :arguments '(array string))
 	(ff:defforeign 'c-glyph :entry-point "GlyphCMNFile" :arg-checking nil :arguments '(fixnum))
 	(ff:defforeign 'c-font-glyph :entry-point "FontGlyphCMNFile" :arg-checking nil :arguments '(string fixnum fixnum))
 	(ff:defforeign 'c-text :entry-point "FontTextCMNFile" :arg-checking nil :arguments '(string fixnum string))
 	(ff:defforeign 'c-stem :entry-point "StemCMNFile" :arg-checking nil :arguments '(array))
 	(ff:defforeign 'c-just-stem :entry-point "VerticalLineCMNFile" :arg-checking nil :arguments '(array))
 	(ff:defforeign 'c-line :entry-point "HorizontalLineCMNFile" :arg-checking nil :arguments '(array))
 	(ff:defforeign 'c-slanted-line :entry-point "SlantedLineCMNFile" :arg-checking nil :arguments '(array))
 	)

#-Excl (progn
	 ;; these are just place holders to make the compiler/loader happy
	 ;; they are not actually called anywhere.

	 (defun c-open (&rest args) (declare (ignore args)) (error "accidental c call") nil)
	 (defun c-close (&rest args) (declare (ignore args)) (error "accidental c call") nil)
	 (defun c-print (&rest args) (declare (ignore args)) (error "accidental c call") nil)
	 (defun c-lineto (&rest args) (declare (ignore args)) (error "accidental c call") nil)
	 (defun c-rlineto (&rest args) (declare (ignore args)) (error "accidental c call") nil)
	 (defun c-moveto (&rest args) (declare (ignore args)) (error "accidental c call") nil)
	 (defun c-rmoveto (&rest args) (declare (ignore args)) (error "accidental c call") nil)
	 (defun c-curveto (&rest args) (declare (ignore args)) (error "accidental c call") nil)
	 (defun c-fill-in (&rest args) (declare (ignore args)) (error "accidental c call") nil)
	 (defun c-glyph (&rest args) (declare (ignore args)) (error "accidental c call") nil)
	 (defun c-font-glyph (&rest args) (declare (ignore args)) (error "accidental c call") nil)
	 (defun c-text (&rest args) (declare (ignore args)) (error "accidental c call") nil)
	 (defun c-stem (&rest args) (declare (ignore args)) (error "accidental c call") nil)
	 (defun c-just-stem (&rest args) (declare (ignore args)) (error "accidental c call") nil)
	 (defun c-line (&rest args) (declare (ignore args)) (error "accidental c call") nil)
	 (defun c-slanted-line (&rest args) (declare (ignore args)) (error "accidental c call") nil)
	 )



(defvar cmn-debug nil)
(defvar *cmn-score* nil)		;for debugging convenience

(defun divide (a b) (if (zerop b) (progn (warn "attempt to divide by 0") a) (/ a b)))


;;;
;;; ----------------    bounds (the "bounding-box")
;;;
;;; x0 y0 is lower left corner, normally -- some objects slop over one way or the other

(defclass write-protect ()
  ())

;;; our basic score syntax includes such things as (cmn staff treble c4 q) and we have
;;; to make sure we never write the fields of the "staff" or "treble" (etc) variables
;;; so all such special variables are "write-protected".

;;; using typep as in:
;;;(defun write-protected (object) (typep object 'write-protect))
;;; is about 10 times slower than methods such as:

(defmethod write-protected ((obj t)) nil)
(defmethod write-protected ((obj write-protect)) t)

(defclass box-mixin ()
  ((x0 :initarg :x0 :initform 0 :reader x0 :reader box-x0 :type number)
   (y0 :initarg :y0 :initform 0 :reader y0 :reader box-y0 :type number)
   (x1 :initarg :x1 :initform 0 :reader x1 :reader box-x1 :type number)
   (y1 :initarg :y1 :initform 0 :reader y1 :reader box-y1 :type number)))

(defclass box (box-mixin)
  ((x0 :accessor x0 :accessor box-x0 :type number)
   (y0 :accessor y0 :accessor box-y0 :type number)
   (x1 :accessor x1 :accessor box-x1 :type number)
   (y1 :accessor y1 :accessor box-y1 :type number)))

(defun copy-bounding-box (obj1 obj2)
  (setf (box-x0 obj2) (box-x0 obj1))
  (setf (box-x1 obj2) (box-x1 obj1))
  (setf (box-y0 obj2) (box-y0 obj1))
  (setf (box-y1 obj2) (box-y1 obj1)))

(defmethod staff-y0 ((box box-mixin)) (box-y0 box))

(defgeneric copy (read-object &optional write-object))
(defgeneric descry (object &optional stream controller))
(defgeneric display (object container score &rest rest))
(defgeneric house (object score))
(defgeneric notify (controller &optional objects))
(defgeneric identify (object))

;;; these are the primary generic functions that every score object should respond to
;;; copy returns a copy (unwrite-protected) of its argument (at top level)
;;; descry is our describe method
;;; other generic functions that lots of objects know about are display, house, store, and notify
;;; display displays an object
;;; house puts the three layers of white space boxes around an object (during justification)
;;; notify is the syntax checker -- it reads all the arguments collected into an object and regularizes everything
;;; identify is used by the error reporting mechanism to try to give an understandable reference to the current object


(defmethod descry (anything &optional stream controller)
  (declare (ignore controller))
  (format stream " ~A" anything))

(defmethod identify (anything)
  anything)




(defmethod copy ((box box-mixin) &optional object)
  (let ((new-box (if (not object) (make-box)
		   (if (write-protected object) (copy object)
		     object))))
    (setf (box-x0 new-box) (box-x0 box))
    (setf (box-y0 new-box) (box-y0 box))
    (setf (box-x1 new-box) (box-x1 box))
    (setf (box-y1 new-box) (box-y1 box))
    (if (next-method-p) (call-next-method box new-box))
    new-box))

(defun not-rational (x)			;flush lisp rationals which server can't handle and prettify print-out
					;~1,3F is needed because kcl prints out endless bogus digits, thereby confusing postscript
					;not-rational should only be used with ~A (sigh -- this is kcl's fault)
  (if (integerp x) x (format nil "~1,3F" (float x))))

(defparameter prewhitespace "                    ")

(defmethod descry ((box box-mixin) &optional stream controller)
  (format stream "~A~A~A~A~A~A~A"
	  (if (not controller) "(box" 
	    (if (plusp (x0 box)) (format nil "~%~A" prewhitespace)
	      ""))
	  (if (not (zerop (x0 box))) (format nil " :x0 ~A" (not-rational (x0 box))) "")
	  (if (not (zerop (y0 box))) (format nil " :y0 ~A" (not-rational (y0 box))) "")
	  (if (not (zerop (x1 box))) (format nil " :x1 ~A" (not-rational (x1 box))) "")
	  (if (not (zerop (y1 box))) (format nil " :y1 ~A" (not-rational (y1 box))) "")
	  (if (next-method-p) (call-next-method box stream (or controller box)) "")
	  (if (not controller) ")" "")))




;;; 
;;; ---------------     patterns and coordinate transformations
;;;
;;; this is not directly a part of the CMN world (it's just graphics fooling around)
;;; so it is not supported in the fanciest possible manner -- upside down bass clef for Bach.
;;; (see transform-box for code to do this silliness correctly)


(defclass matrix-mixin ()
  ((matrix :initarg :matrix 
	   :initform nil
	   :reader matrix)))

(defclass matrix (matrix-mixin)
  ((matrix :accessor matrix)))

(defmethod matrix-p ((obj t)) nil)
(defmethod matrix-p ((obj matrix-mixin)) t)

(defmethod copy ((matrix matrix-mixin) &optional object)
  (let ((new-matrix (if (not object) (make-matrix)
		      (if (write-protected object) (copy object)
			object))))
    (if (matrix matrix) (setf (matrix new-matrix) (copy-list (matrix matrix))))
    (if (next-method-p) (call-next-method matrix new-matrix))
    new-matrix))

(defun identity-matrix-p (object)
  (let ((mat (matrix object)))
    (or (not mat)
	(equal mat '(1 0 0 1 0 0)))))

(defun invisible-matrix-p (object)
  (equal (matrix object) '(0 0 0 0 0 0)))

(defmethod descry ((matrix matrix-mixin) &optional stream controller)
  (format stream "~A~A~A~A"
	  (if (not controller) "(matrix" 
	    (if (not (identity-matrix-p matrix))
		(format nil "~%~A" prewhitespace)
	      ""))
	  (if (not (identity-matrix-p matrix))
	      (format nil " :matrix '~A"
		      (map 'list #'not-rational (matrix matrix)))
	    "")
	  (if (next-method-p) (call-next-method matrix stream (or controller matrix)) "")
	  (if (not controller) ")" "")))

(defun identify-matrix (object)
  (if (not (identity-matrix-p object))
      (if (invisible-matrix-p object)
	  " (scale 0 0)"
	(if (equal (matrix object) '(-1 0 0 1 0 0))
	    " mirrored"
	  (if (let ((mobj (matrix object)))
		(and (zerop (second mobj)) (zerop (third mobj)) (zerop (fifth mobj)) (zerop (sixth mobj))))
	      (format nil " (scale ~1,3F ~1,3F)" (first (matrix object)) (fourth (matrix object)))
	    (format nil " (matrix '~A)" (map 'list #'not-rational (matrix object))))))
    ""))



;;; patterns (at this level) are functions that are wrapped around the thing being displayed in place of the "show" or "stroke" command

(defclass pattern-mixin ()
  ((pattern :initarg :pattern :initform nil :reader pattern)))

(defclass pattern (pattern-mixin)
  ((pattern :accessor pattern)))

(defmethod copy ((pattern pattern-mixin) &optional object)
  (let ((new-pattern (if (not object) (make-pattern)
		       (if (write-protected object) (copy object)
			 object))))
    (if (pattern pattern) (setf (pattern new-pattern) (copy-list (pattern pattern))))
    (if (next-method-p) (call-next-method pattern new-pattern))
    new-pattern))

(defmethod descry ((pattern pattern-mixin) &optional stream controller)
  (format stream "~A~A~A~A"
	  (if (not controller) "(pattern" 
	    (if (pattern pattern) (format nil "~%~A" prewhitespace)
	      ""))
	  (if (pattern pattern)
	      (format nil " :pattern ~A" (pattern pattern))
	    "")
	  (if (next-method-p) (call-next-method pattern stream (or controller pattern)) "")
	  (if (not controller) ")" "")))

(defvar pends nil)

(defmethod display :before ((any-object pattern-mixin) container score &rest rest)
  (declare (ignore container))
  (when (and (pattern any-object) (not (member :just-fooling rest)))
    (multiple-value-bind
	(pstart tpend pend)
	(funcall (pattern any-object) score any-object)
      (declare (ignore tpend))
      (push pend pends)
      (if (not (c-output-in-progress score))
	  (format (output score) " gsave ~A~%" pstart)
	(c-print (format nil " gsave ~A" pstart))))))

(defmethod display :after ((any-object pattern-mixin) container score &rest rest)
  (declare (ignore container))
  (when (and (pattern any-object) (not (member :just-fooling rest)))
    (if (not (c-output-in-progress score))
	(format (output score) " ~A grestore~%" (pop pends))
      (c-print (format nil " ~A grestore" (pop pends))))))




;;; maxtrix multiplication done by hand (no need for full multiply) 
;;; (third column is assumed to be 0 0 1 since we're in flatland)
;;; rather than optimize using rotate scale concat etc, I'll just use concat.

(defun cosd (n) (cos (* n (/ pi 180))))
(defun sind (n) (sin (* n (/ pi 180))))

(defun %rotate-matrix (matrix angle)	;degrees
  (let ((cs (coerce (cosd angle) 'short-float))
	(sn (coerce (sind angle) 'short-float)))
    (if (null matrix)
	(list cs sn (- sn) cs 0 0)
      (let ((a (first matrix))
	    (b (second matrix))
	    (c (third matrix))
	    (d (fourth matrix))
	    (e (fifth matrix))
	    (f (sixth matrix)))
	(list (+ (* a cs) (* c sn))
	      (+ (* b cs) (* d sn))
	      (- (* c cs) (* a sn))
	      (- (* d cs) (* b sn))
	      e f)))))

(defun %scale-matrix (matrix x-scale y-scale)
  (if (not matrix)
      (list x-scale 0 0 y-scale 0 0)
    (list (* x-scale (first matrix))
	  (* x-scale (second matrix))
	  (* y-scale (third matrix))
	  (* y-scale (fourth matrix))
	  (fifth matrix)
	  (sixth matrix))))

(defun %mirror-matrix (matrix)		;could be scale with -x, of course
  (if matrix
      (append (list (- (first matrix))
		    (- (second matrix)))
	      (cddr matrix))
    (list -1 0 0 1 0 0)))

(defun %truncated-matrix-multiply (m1 m2)
  (let ((a1 (first m1))  (a2 (first m2))
	(b1 (second m1)) (b2 (second m2))
	(c1 (third m1))  (c2 (third m2))
	(d1 (fourth m1)) (d2 (fourth m2))
	(e1 (fifth m1))  (e2 (fifth m2))
	(f1 (sixth m1))  (f2 (sixth m2)))
    (list (+ (* a1 a2) (* b1 c2))
	  (+ (* a1 b2) (* b1 d2))
	  (+ (* c1 a2) (* d1 c2))
	  (+ (* c1 b2) (* d1 d2))
	  (+ (* e1 a2) (* f1 c2) e2)
	  (+ (* e1 b2) (* f1 d2) f2))))


(defun transform-box (mat tv bb)
  
  ;;Transforming Axis-Aligned Bounding Boxes
  ;;by Jim Arvo
  ;;from "Graphics Gems", Academic Press, 1990
  ;;
  ;;Transforms a 3D axis-aligned box via a 3x3 matrix and a translation
  ;;vector and returns an axis-aligned box enclosing the result.
  ;;
  ;;reduced here to 2-D and reversed sense of rotation angle
  ;;(original was clockwise, but PostScript is counterclockwise) 
  
  (let* ((amin (make-array 2 :element-type 'short-float))
	 (amax (make-array 2 :element-type 'short-float))
	 (bmin (make-array 2 :element-type 'short-float))
	 (bmax (make-array 2 :element-type 'short-float))
	 (a 0.0) (b 0.0))
    
    (setf (aref amin 0) (float (first bb)))
    (setf (aref amin 1) (float (second bb)))
    
    (setf (aref amax 0) (float (third bb)))
    (setf (aref amax 1) (float (fourth bb)))
    
    (setf (aref bmin 0) (float (first tv)))
    (setf (aref bmin 1) (float (second tv)))
    
    (setf (aref bmax 0) (float (first tv)))
    (setf (aref bmax 1) (float (second tv)))
    
    (loop for i from 0 to 1 do
      (loop for j from 0 to 1 do
	(setf a (* (aref amin j) (nth (+ (* j 2) i) mat)))
	(setf b (* (aref amax j) (nth (+ (* j 2) i) mat)))
	(if (< a b)
	    (progn
	      (incf (aref bmin i) a)
	      (incf (aref bmax i) b))
	  (progn
	    (incf (aref bmin i) b)
	    (incf (aref bmax i) a)))))
    (list (aref bmin 0) (aref bmin 1)
	  (aref bmax 0) (aref bmax 1))))




;;;
;;; ----------------    glyphs (font characters)
;;;

(defclass write-protected-marks-mixin ()
  ((marks :initarg :marks :initform nil :reader marks)))

(defclass marks-mixin (write-protected-marks-mixin)
  ((marks :accessor marks)))

(defmethod marks-p ((obj t)) nil)
(defmethod marks-p ((obj marks-mixin)) t)

(defmethod descry ((vis write-protected-marks-mixin) &optional stream controller)
  (format stream "~A~A"
	  (if (marks vis) 
	      (format stream " :marks (list~{ ~A~})"
		      (loop for mark in (marks vis) collect (descry mark stream nil)))
	    "")
	  (if (next-method-p) (call-next-method vis stream (or controller vis)) "")))

(defmethod copy ((vis write-protected-marks-mixin) &optional object)
  (let ((new-vis (if (not object) (make-marks-mixin)
		   (if (write-protected object) (copy object)
		     object))))
    (if (marks vis) (setf (marks new-vis) (loop for mark in (marks vis) collect (copy mark))))
    (if (next-method-p) (call-next-method vis new-vis))
    new-vis))

(defun display-marks (object score &rest rest)
  (if (marks object) 
      (loop for mark in (marks object) do 
	(if (or (not (sundry-p mark))
		(not (eq (sundry-name mark) :slur)))
	    (apply #'display mark object score rest)))))

(defmethod add-to-marks ((anything write-protected-marks-mixin) objects) 
  (setf (marks anything) (append (marks anything) objects)))

(defun edit-mark (anything mark-identifier mark-editor)
  (loop for m in (marks anything) do
    (if (funcall mark-identifier m anything)
	(funcall mark-editor m anything))))

(defun identify-marks (object &optional file)
  (if (marks object)
      (format file "~{ ~A~}" 
	      (loop for mark in (marks object)
	       collect (identify mark)))
    ""))


(defclass visible-mixin (write-protected-marks-mixin box-mixin pattern-mixin matrix-mixin)
  ((justification :initform nil :initarg :justification :reader visible-justification)))

(defclass visible (visible-mixin marks-mixin box pattern matrix)
  ((justification :accessor visible-justification :accessor justification)))

(defmethod visible-p ((obj t)) nil)
(defmethod visible-p ((obj visible-mixin)) t)

(defmethod descry ((vis visible-mixin) &optional stream controller)
  (format stream "~A~A"
	  (if (visible-justification vis) 
	      (format stream " :justification :~(~A~)" (visible-justification vis))
	    "")
	  (if (next-method-p) (call-next-method vis stream (or controller vis)) "")))

(defmethod copy ((vis visible-mixin) &optional object)
  (let ((new-vis (if (not object) 
		     (cmn-error "attempt to copy a bare instance of the visible-mixin class")
		   (if (write-protected object) (copy object)
		     object))))
    (setf (visible-justification new-vis) (visible-justification vis) )
    (if (next-method-p) (call-next-method vis new-vis))
    new-vis))

(defun identify-visible (object &optional file)
  (if (visible-justification object)
      (if (eq (visible-justification object) :none)
	  (format file " unjustified")
	(format file " (justification :~(~A~)" (visible-justification object)))
    ""))



(defclass glyph-mixin (visible-mixin)
  ((rx :initarg :rx :initform 0 :reader rx :reader g-rx)
   (index :initarg :index :initform nil :reader index)))

(defclass glyph (glyph-mixin visible)
  ((rx :accessor rx)
   (index :accessor index)))

(defclass write-protected-glyph (write-protect glyph-mixin)
  ())

(defmethod glyph-p ((obj t)) nil)
(defmethod glyph-p ((obj glyph-mixin)) t)

(defmethod copy ((glyph glyph-mixin) &optional object)
  (let ((new-glyph (if (not object) (make-glyph)
		     (if (write-protected object) (copy object)
		       object))))
    (setf (rx new-glyph) (g-rx glyph))
    (setf (index new-glyph) (index glyph))
    (if (next-method-p) (call-next-method glyph new-glyph))
    new-glyph))

(defmethod descry ((glyph glyph-mixin) &optional stream controller)
  (format stream "~A~A~A~A"
	  (if (not controller) "(glyph" 
	    (if (index glyph) (format nil "~%~A" prewhitespace)
	      ""))
	  (if (index glyph)
	      (format nil " :index ~A :rx ~A" (index glyph) (not-rational (rx glyph)))
	    "")
	  (if (next-method-p) (call-next-method glyph stream (or controller glyph)) "")
	  (if (not controller) ")" "")))

(defun glyph (&rest args)
  (let ((glf (make-glyph))) 
    (loop for act in args do
      (when act
	(if (glyph-p act)
	    (progn
	      (setf (index glf) (index act))
	      (setf (rx glf) (rx act)))
	  (if (self-acting-p act)
	      (funcall (action act) glf (arguments act))
	    (if (integerp act)
		(setf (index glf) act)
	      (if (characterp act)
		  (setf (index glf) (decimal-to-octal (char-code act)))))))))
    glf))



(defclass font-mixin ()
  ((font-size :initarg :font-size :initform nil :accessor font-size)
   (font-name :initarg :font-name :initform nil :accessor font-name)
   (font-scaler :initarg :font-scaler :initform 1.0 :accessor font-scaler)))

(defmethod descry ((fnt font-mixin) &optional stream controller)
  (format stream "~A~A~A~A~A~A"
	  (if (not controller) "(font" "")
	  (if (font-name fnt) (format nil " :font-name ~A" (font-name fnt)) "")
	  (if (font-size fnt) (format nil " :font-size ~A" (font-size fnt)) "")
	  (if (/= (font-scaler fnt) 1.0) (format nil " :font-scaler ~A" (font-scaler fnt)) "")
	  (if (next-method-p) (call-next-method fnt stream (or controller fnt)) "")
	  (if (not controller) ")" "")))

(defmethod copy ((fnt font-mixin) &optional object)
  (let ((new-fnt (if (not object) (make-font-mixin)
		   (if (write-protected object) (copy object)
		     object))))
    (setf (font-size new-fnt) (font-size fnt))
    (setf (font-name new-fnt) (font-name fnt))
    (setf (font-scaler new-fnt) (font-scaler fnt))
    (if (next-method-p) (call-next-method fnt new-fnt))
    new-fnt))




(defclass dxy-mixin ()
  ((dx :initarg :dx :initform 0 :reader dx :reader dxy-dx)
   (dy :initarg :dy :initform 0 :reader dy :reader dxy-dy)))

(defclass dxy (dxy-mixin)
  ((dx :accessor dx :accessor dxy-dx)
   (dy :accessor dy :accessor dxy-dy)))

(defmethod dxy-p ((obj t)) nil)
(defmethod dxy-p ((obj dxy-mixin)) t)

(defmethod descry ((dxy dxy-mixin) &optional stream controller)
  (format stream "~A~A~A" 
	  (if (not (zerop (dx dxy))) (format nil " :dx ~1,3F" (dx dxy)) "")
	  (if (dy dxy)
	      (if (listp (dy dxy))
		  (format nil " :dy '~A " (dy dxy))
		(if (not (zerop (dy dxy))) 
		    (format nil " :dy ~1,3F" (dy dxy)) 
		  ""))
	    "")
	  (if (next-method-p) (call-next-method dxy stream (or controller dxy)) "")))

(defmethod copy ((dxy dxy-mixin) &optional object)
  (let ((new-dxy (if (not object) (make-dxy-mixin)
		   (if (write-protected object) (copy object)
		     object))))
    (setf (dxy-dx new-dxy) (dxy-dx dxy))
    (setf (dxy-dy new-dxy) (dxy-dy dxy))
    (if (next-method-p) (call-next-method dxy new-dxy))
    new-dxy))

(defun identify-dxy (object)
  (if (or (not (zerop (dx object)))
	  (not (zerop (dy object))))
      (format nil "~A~A" 
	      (if (not (zerop (dx object))) (format nil " (dx ~1,3F)" (dx object)) "")
	      (if (not (zerop (dy object))) (format nil " (dy ~1,3F)" (dy object)) ""))
    ""))


(defclass text (dxy font-mixin visible)
  ((letters :initarg :letters :initform nil :accessor letters)
   (y :initarg :y :initform nil :accessor y :accessor text-y)
   (x :initarg :x :initform nil :accessor x :accessor text-x)))

(defmethod text-p ((obj t)) nil)
(defmethod text-p ((obj text)) t)

(defmethod descry ((text text) &optional stream controller)
  (format stream "~A~A~A~A"
	  (if (not controller) "(text" "")
	  (if (letters text) (format nil " :letters ~A" (letters text)) "")
	  (if (next-method-p) (call-next-method text stream (or controller text)) "")
	  (if (not controller) ")" "")))

(defmethod copy ((text text) &optional object)
  (let ((new-text (if (not object) (make-text)
		    (if (write-protected object) (copy object)
		      object))))
    (setf (letters new-text) (letters text))
    (setf (text-x new-text) (text-x text))
    (setf (text-y new-text) (text-y text))
    (if (next-method-p) (call-next-method text new-text))
    new-text))




(defgeneric %rotate (object angle))
(defgeneric %scale (object xscl yscl))
(defgeneric %mirror (object &rest rest) )
(defgeneric %transform (object matrix))

(defmethod %rotate ((object matrix) angle)
  (setf (matrix object) (%rotate-matrix (or (matrix object) '(1 0 0 1 0 0)) angle))
  object)

(defmethod %scale ((object matrix) xscl yscl)
  (setf (matrix object) (%scale-matrix (matrix object) (or xscl 1) (or yscl 1)))
  object)

(defmethod %scale-1 ((object matrix) scls)
  (setf (matrix object) (%scale-matrix (matrix object) (or (first scls) 1) (or (second scls) 1)))
  object)

(defmethod %mirror ((object matrix) &rest rest)
  (declare (ignore rest))
  (setf (matrix object) (%mirror-matrix (matrix object)))
  object)

(defmethod %transform ((object matrix) matrix)
  (setf (matrix object) (%truncated-matrix-multiply (or (matrix object) '(1 0 0 1 0 0)) matrix))
  object)

(defmethod %rotate ((object write-protect) angle)
  (let ((new-object (copy object)))
    (setf (matrix new-object) (%rotate-matrix (or (matrix new-object) '(1 0 0 1 0 0)) angle))
    new-object))

(defmethod %scale ((object write-protect) xscl yscl)
  (let ((new-object (copy object)))
    (setf (matrix new-object) (%scale-matrix (matrix new-object) (or xscl 1) (or yscl 1)))
    new-object))

(defmethod %mirror ((object write-protect) &rest rest)
  (declare (ignore rest))
  (let ((new-object (copy object)))
    (setf (matrix new-object) (%mirror-matrix (matrix new-object)))
    new-object))

(defmethod %transform ((object write-protect) matrix)
  (let ((new-object (copy object)))
    (%truncated-matrix-multiply (or (matrix new-object) '(1 0 0 1 0 0))  matrix)
    new-object))



;;;
;;; ----------------    the graphics server
;;;
;;;   lineto moveto rlineto rmoveto show
;;;   initialize header footer finalize
;;;   comment line-width draw fill-in
;;;   rotate scale mirror 
;;;   x y (current drawing position)

(defgeneric lineto (server x y))	;draw a line to (x y)
(defgeneric rlineto (server x y))	;draw a line to (current-x + x, current-y + y)
(defgeneric moveto (server x y))	;hop to (x y)
(defgeneric rmoveto (server x y))	;hop to (current-x + x, current-y + y)
(defgeneric curveto (server x0 y0 x1 y1 x2 y2)) ;draw a curve 
(defgeneric show (server x &key))	;cause "x" to be displayed
(defgeneric initialize (server &key))	;prepare a clean slate
(defgeneric header (server &key))	;whatever setup commands are needed by the server
(defgeneric footer (server &key))	;whatever shut-down commands are needed
(defgeneric finalize (server &key))	;finish graphics
(defgeneric comment (server x))		;put a comment in the server command stream (optional)
(defgeneric draw (server &key))		;draw the current path
(defgeneric fill-in (server &key))	;fill the current object

;;;
;;; it is my belief that all the system-specific graphics code is in this file, and the other files
;;; build on whatever is here, so only this file needs to be changed to move to a different 
;;; graphics environment.  Everything here is currently PostScript oriented, but the only
;;; odd things are the special matrix/pattern stuff (which QuickDraw expects you to do on 
;;; the bitmaps, I guess), and Bezier curves used for ties and slurs. See the function
;;; bezier to draw the latter using just lineto's.
;;;
;;; (There are two additional format statements in cmn4 in the display functions toward the end).
;;; (There is also the nxtcmn.c file of format optimizations)

(defvar *cmn-score-size* nil)
(defvar *old-cmn-score-size* nil)

(defclass server (box-mixin matrix)
  ((x :initarg :x :initform nil :accessor x :accessor scr-x) ;current drawing position (known as the turtle position in earlier systems)
   (y :initarg :y :initform nil :accessor y :accessor scr-y)
   (size :initarg :size :accessor size :accessor scr-size :initform 1)
   (scaling-matrix :initarg :scaling-matrix :accessor scaling-matrix :initform nil)
   (output :initform nil :initarg :output :accessor output) ;output command stream
   (c-output-in-progress :initform nil :accessor c-output-in-progress)
   ;; this slot is t if C format calls are in progress (much faster than Lisp's format)
   (output-file :initform nil :initarg :output-file :accessor output-file)
   (line-width :initform 0 :initarg :line-width :accessor line-width)
   (font :initform nil :initarg :font :accessor font)
   (prologue :initarg :prologue :initform nil :accessor prologue)
   (bounded :initform nil :initarg :bounded :accessor bounded) ;is server supposed to track the excursion religiously.
   (x0 :accessor scr-x0 :type number)	; in ACL, by giving a unique name to the accessor, slot accesses are sped up by a factor of 10
   (y0 :accessor scr-y0 :type number)
   (x1 :accessor scr-x1 :type number)
   (y1 :accessor scr-y1 :type number)))


(defmethod copy ((server server) &optional object)
  (declare (ignore object))
  (error "can't copy ~A" server))

(defmethod descry ((server server) &optional stream controller)
  (format stream "~A~A~A~A~A~A~A~A~A~A"
	  (if (not controller) "(server" 
	    (format nil "~%    "))
	  (if (x server) (format nil " :x ~A" (x server)) "")
	  (if (y server) (format nil " :y ~A" (y server)) "")
	  (if (size server) (format nil " :size ~A" (size server)) "")
	  (if (output server) (format nil " :output-file ~A" (output-file server)) "")
	  (if (not (zerop (line-width server))) (format nil " :line-width ~A" (line-width server)) "")
	  (if (font server) (format nil " :font ~A" (font server)) "")
	  (if (bounded server) (format nil " :bounded ~A" (bounded server)) "")
	  (if (next-method-p) (call-next-method server stream (or controller server)) "")
	  (if (not controller) ")" "")))

(defmethod clear-bounding-box ((server server))
  (setf (scr-x0 server) 1000)
  (setf (scr-x1 server) 0)
  (setf (scr-y0 server) 1000)
  (setf (scr-y1 server) 0))

(defgeneric bounds (server))
(defgeneric (setf bounds) (box server))

(defmethod %update-bounds ((server server))
  (declare (optimize (speed 3) (safety 0)))
  (setf (scr-x0 server) (min (scr-x0 server) (scr-x server)))
  (setf (scr-y0 server) (min (scr-y0 server) (scr-y server)))
  (setf (scr-x1 server) (max (scr-x1 server) (scr-x server)))
  (setf (scr-y1 server) (max (scr-y1 server) (scr-y server))))

(defmethod %xy ((server server) x y) 
  (setf (scr-x server) x)
  (setf (scr-y server) y))

(defmethod %dxy ((server server) dx dy)
  (incf (scr-x server) dx)
  (incf (scr-y server) dy))

(defmethod bounds ((server server))
  (list (scr-x0 server) (scr-y0 server) (scr-x1 server) (scr-y1 server)))

(defmethod (setf bounds) (box (server server)) ;default here is a list passed with box coordinates
  (declare (optimize (speed 3) (safety 0)))
  (when (not (bounded server))
    (setf (scr-x0 server) (min (scr-x0 server) (first box)))
    (setf (scr-y0 server) (min (scr-y0 server) (second box)))
    (setf (scr-x1 server) (max (scr-x1 server) (third box)))
    (setf (scr-y1 server) (max (scr-y1 server) (fourth box)))))

(defmethod (setf bounds) ((box box) (server server))
  (setf (bounds server) (list (box-x0 box) (box-y0 box) (box-x1 box) (box-y1 box))))

(defvar c-args (make-array 6 :element-type 'short-float :initial-element 0.0))

(defmethod lineto ((server server) x y)
  (declare (optimize (speed 3) (safety 0)))
  (let* ((size (or *cmn-score-size* (scr-size server)))
	 (xs (float (* x size)))
	 (ys (float (* y size))))
    (if (output server)	
	(if (not (c-output-in-progress server))
	    (format (output server) " ~1,2F ~1,2F lineto~%" xs ys)
	  (progn 
	    (setf (aref c-args 0) xs) 
	    (setf (aref c-args 1) ys)
	    (c-lineto c-args))))
    (setf (scr-x server) xs)
    (setf (scr-y server) ys)
    (setf (scr-x0 server) (min (scr-x0 server) xs))
    (setf (scr-y0 server) (min (scr-y0 server) ys))
    (setf (scr-x1 server) (max (scr-x1 server) xs))
    (setf (scr-y1 server) (max (scr-y1 server) ys))))

(defmethod rlineto ((server server) dx dy)
  (declare (optimize (speed 3) (safety 0)))
  (let* ((size (or *cmn-score-size* (scr-size server)))
	 (dxs (float (* dx size)))
	 (dys (float (* dy size))))
    (if (output server)	
	(if (not (c-output-in-progress server))
	    (format (output server) " ~1,2F ~1,2F rlineto~%" dxs dys)	    
	  (progn 
	    (setf (aref c-args 0) dxs) 
	    (setf (aref c-args 1) dys)
	    (c-rlineto c-args))))
    (setf dxs (incf (scr-x server) dxs))
    (setf dys (incf (scr-y server) dys))
    (setf (scr-x0 server) (min (scr-x0 server) dxs))
    (setf (scr-y0 server) (min (scr-y0 server) dys))
    (setf (scr-x1 server) (max (scr-x1 server) dxs))
    (setf (scr-y1 server) (max (scr-y1 server) dys))))

(defmethod moveto ((server server) x y)
  ;; most of the run time is spent here -- all in Lisp's format function
  (declare (optimize (speed 3) (safety 0)))
  (let* ((size (or *cmn-score-size* (scr-size server)))
	 (xs (float (* x size)))
	 (ys (float (* y size))))
    (if (output server)	
	(if (not (c-output-in-progress server))
	    (format (output server) " ~1,2F ~1,2F moveto~%" xs ys)
	  (progn 
	    (setf (aref c-args 0) xs) 
	    (setf (aref c-args 1) ys)
	    (c-moveto c-args))))
    (setf (scr-x server) xs)
    (setf (scr-y server) ys)
    (setf (scr-x0 server) (min (scr-x0 server) xs))
    (setf (scr-y0 server) (min (scr-y0 server) ys))
    (setf (scr-x1 server) (max (scr-x1 server) xs))
    (setf (scr-y1 server) (max (scr-y1 server) ys))))

(defmethod rmoveto ((server server) dx dy)
  (declare (optimize (speed 3) (safety 0)))
  (let* ((size (or *cmn-score-size* (scr-size server)))
	 (dxs (float (* dx size)))
	 (dys (float (* dy size))))
    (if (output server) 
	(if (not (c-output-in-progress server))
	    (format (output server) " ~1,2F ~1,2F rmoveto~%" dxs dys)
	  (progn 
	    (setf (aref c-args 0) dxs) 
	    (setf (aref c-args 1) dys)
	    (c-rmoveto c-args))))
    (setf dxs (incf (scr-x server) dxs))
    (setf dys (incf (scr-y server) dys))
    (setf (scr-x0 server) (min (scr-x0 server) dxs))
    (setf (scr-y0 server) (min (scr-y0 server) dys))
    (setf (scr-x1 server) (max (scr-x1 server) dxs))
    (setf (scr-y1 server) (max (scr-y1 server) dys))))

(defmethod comment ((server server) x)
  (if (output server) 
      (if (not (c-output-in-progress server))
	  (format (output server) "% ~A~%" x)
	(c-print (format nil "% ~A" x)))))

(defmethod (setf line-width) :after (x (server server))
  (if (output server) 
      (if (not (c-output-in-progress server))
	  (format (output server) " ~A setlinewidth~%" (not-rational (* x (or *cmn-score-size* (scr-size server)))))
	(if (zerop x)
	    (c-print " 0 setlinewidth")
	  (c-print (format nil " ~A setlinewidth" (not-rational (* x (or *cmn-score-size* (scr-size server))))))))))

(defmethod draw ((server server) &key pattern &allow-other-keys)
  (if (output server)
      (if (not (c-output-in-progress server))
	  (if (not pattern)
	      (format (output server) " stroke~%")
	    (progn
	      ;; pattern here is some list of numbers being dash-len space-len or whatever
	      (format (output server) " [~{~A ~}] 0 setdash stroke~%" (map 'list #'not-rational pattern))
	      (format (output server) " [] 0 setdash~%")))
	(if (not pattern)
	    (c-print " stroke")
	  (c-print (format nil " [~{~A ~}] 0 setdash stroke [] 0 setdash" (map 'list #'not-rational pattern)))))))


(defmethod simple-draw ((server server))
  (if (output server) 
      (if (not (c-output-in-progress server))
	  (format (output server) " stroke~%")
	(c-print " stroke"))))


(defgeneric %rx (server object &key) )

(defmethod %rx ((server server) (glyph glyph-mixin) &key (count 1) size &allow-other-keys)
  (if (identity-matrix-p glyph)
      (* (g-rx glyph) count (or size *cmn-score-size* (scr-size server)))
    0))					;if matrix then we used gsave/grestore and the outer point did not move

(defmethod simple-show ((server server) glf)
  ;; this is only used when we know what the glf is (as write-protected) and nothing is happening
  (when (output server)
    (if (not (c-output-in-progress server))
	(princ (format nil  " (\\~D) show~%" (index glf)) (output server))
      (c-glyph (index glf))))
  (let ((dxs (* (g-rx glf) (or *cmn-score-size* (scr-size server)))))
    (setf dxs (incf (scr-x server) dxs))
    (setf (scr-x0 server) (min (scr-x0 server) dxs))
    (setf (scr-x1 server) (max (scr-x1 server) dxs))))

(defmethod show ((server server) (glyph glyph-mixin) &key (count 1) size &allow-other-keys)
  (when (not (invisible-matrix-p glyph))
    (let ((all-done nil))
      (when (output server)
	(let ((c-out (c-output-in-progress server))
	      (pat (pattern glyph))
	      (no-mat (identity-matrix-p glyph)))
	  (if (and c-out no-mat (= count 1) (not pat)) ;optimize the most common cases
	      (if size
		  (c-font-glyph (font server) (floor size) (index glyph))
		(progn
		  (setf all-done t)
		  (simple-show server glyph)))
					;		(c-glyph (index glyph))
	    (let ((glyphs (loop for i from 1 to count collect (index glyph)))
		  (pstart "")
		  (pend "show"))
	      (when size 
		(if (not c-out)
		    (format (output server) " gsave~% /~A findfont ~D scalefont setfont~%" (font server) size)
		  (c-print (format nil " gsave~% /~A findfont ~D scalefont setfont" (font server) size))))
	      (if pat
		  (multiple-value-bind
		      (qpstart qpend) 
		      (funcall pat server glyph)
		    (if qpstart (setf pstart qpstart))
		    (if qpend (setf pend qpend))))
	      (if no-mat
		  (if (not c-out)
		      (princ (format nil " ~A(~{\\~D~}) ~A~A~%" pstart glyphs pend (if size " grestore" "")) (output server))
		    (c-print (format nil " ~A(~{\\~D~}) ~A~A" pstart glyphs pend (if size " grestore" ""))))
		(if (not c-out)
		    (princ (format nil " ~A~A(~{\\~D~}) [ ~{~A ~}] concat ~A grestore~%"
				   (if (not size) "gsave " "")
				   pstart glyphs (map 'list #'not-rational (matrix glyph)) pend)
			   (output server))
		  (c-print (format nil " ~A~A(~{\\~D~}) [ ~{~A ~}] concat ~A grestore"
				   (if (not size) "gsave " "")
				   pstart glyphs (map 'list #'not-rational (matrix glyph)) pend))))))))
      (if (not all-done)
	  (let* ((size-server (or *cmn-score-size* (scr-size server)))
		 (x0g (box-x0 glyph))
		 (x1g (box-x1 glyph))
		 (rxg (g-rx glyph))
		 (xs (scr-x server))
		 (ys (scr-y server))
		 (orig-box (list 0
				 (* size-server (box-y0 glyph))
				 (* (or size size-server) 
				    (* count (if (zerop rxg)
						 (if (< x0g x1g)
						     (- x1g x0g)
						   x1g)
					       rxg)))
				 (* (or size size-server) (box-y1 glyph))))
		 (new-box (if (identity-matrix-p glyph)
			      (list (+ xs (first orig-box))
				    (+ ys (second orig-box))
				    (+ xs (third orig-box))
				    (+ ys (fourth orig-box)))
			    (transform-box (matrix glyph) (list xs ys) orig-box))))
	    (setf (bounds server) new-box)
	    (if (identity-matrix-p glyph) 
		(%dxy server (%rx server glyph :count count :size size) 0)))))))

(defmacro with-scaling (server size x0 y0 &body body)
  `(let ((current-box (bounds ,server))
	 (current-scaling-matrix (scaling-matrix ,server)))
     ;; bounding-box for duration of this scaling is scaled both ways by size and translated by x0 y0
     (clear-bounding-box ,server)
     (setf (scaling-matrix ,server) 
       (%truncated-matrix-multiply (or current-scaling-matrix 
				       (list 1 0 0 1 0 0))
				   (list ,size 0 0 ,size ,x0 ,y0)))
     (if (not (c-output-in-progress ,server))
	 (progn
	   (format (output ,server) " gsave~%")
	   (format (output ,server) " [~1,3F 0 0 ~1,3F ~1,3F ~1,3F] concat~%" ,size ,size ,x0 ,y0))
       (c-print (format nil " gsave~% [~1,3F 0 0 ~1,3F ~1,3F ~1,3F] concat" ,size ,size ,x0 ,y0)))
     ,@body
     ;; I originally had ,.body here, but in ACL that sometimes gets confused and repeats the last statement over and over
     (setf (x0 ,server) (min (first current-box) (+ (* ,size (x0 ,server)) ,x0)))
     (setf (y0 ,server) (min (second current-box) (+ (* ,size (y0 ,server)) ,y0)))
     (setf (x1 ,server) (max (third current-box) (+ (* ,size (x1 ,server)) ,x0)))
     (setf (y1 ,server) (max (fourth current-box) (+ (* ,size (y1 ,server)) ,y0)))
     ;; should we fix up (x/y server) here?
     (setf (scaling-matrix ,server) current-scaling-matrix)
     (if (not (c-output-in-progress ,server))
	 (format (output ,server) " grestore~%")
       (c-print " grestore"))))

(defmacro with-pattern (server pattern &body body)
  `(multiple-value-bind
       (pstart pend)
       (funcall (pattern ,pattern) ,server ,pattern)
     (if (not (c-output-in-progress ,server))
	 (progn
	   (format (output ,server) " gsave~%")
	   (format (output ,server) " ~A~%" pstart))
       (c-print (format nil " gsave~% ~A~%" pstart)))
     ,@body
     (if (not (c-output-in-progress ,server))
	 (format (output ,server) " ~A grestore~%" pend)
       (c-print (format nil " ~A grestore" pend)))))

(defun matrix-front (server matrix)
  (if (and matrix
	   (not (equal matrix '(1 0 0 1 0 0))))
      (if (not (c-output-in-progress server))
	  (format (output server) " gsave~% [ ~{~A ~}] concat~%" (map 'list #'not-rational matrix))
	(c-print (format nil " gsave~% [ ~{~A ~}] concat~%" (map 'list #'not-rational matrix))))))

(defun matrix-back (server matrix)
  (if (and matrix
	   (not (equal matrix '(1 0 0 1 0 0))))
      (if (not (c-output-in-progress server))
	  (format (output server) " grestore~%")
	(c-print " grestore"))))

(defmacro with-transformation (server matrix x0 y0 &body body)
  `(let ((matr (copy-list (if (matrix-p ,matrix) (matrix ,matrix) ,matrix)))
	 (current-box (bounds ,server))
	 (current-scaling-matrix (scaling-matrix ,server)))
     ;; here the problem is similar to the with-scaling case, but we need to transform the
     ;; resultant box by the transforming matrix before accumulating the edges into the box
     (clear-bounding-box ,server)
     (setf (fifth matr) ,x0)
     (setf (sixth matr) ,y0)
     (setf (scaling-matrix ,server) 
       (%truncated-matrix-multiply (or current-scaling-matrix 
				       (list 1 0 0 1 0 0))
				   matr))
     (if (not (c-output-in-progress ,server))
	 (progn
	   (format (output ,server) " gsave~%")
	   (format (output ,server) " [ ~{~A ~}] concat~%" (map 'list #'not-rational matr)))
       (c-print (format nil " gsave~% [ ~{~A ~}] concat~%" (map 'list #'not-rational matr))))
     ,@body
     (let ((new-box (transform-box matr (list ,x0 ,y0) (bounds ,server))))
       (setf (x0 ,server) (min (first current-box) (first new-box)))
       (setf (y0 ,server) (min (second current-box) (second new-box)))
       (setf (x1 ,server) (max (third current-box) (third new-box)))
       (setf (y1 ,server) (max (fourth current-box) (fourth new-box))))
     (setf (scaling-matrix ,server) current-scaling-matrix)
     (if (not (c-output-in-progress ,server))
	 (format (output ,server) " grestore~%")
       (c-print " grestore"))))

;;; in Postscript, if the string to be printed contains "(" or ")" the
;;;   only safe way to deal with them is to prepend the escape character "\"
(defun ps-letters (ur-letters)
  (let ((letters (letters ur-letters)))
    (let ((trouble (or (find #\( letters)
		       (find #\) letters))))
      (if trouble
	  (let* ((baddies (+ (count #\( letters) (count #\) letters)))
		 (inlen (length letters))
		 (i -1)
		 (strlen (+ inlen baddies))
		 (newstr (make-string strlen)))
	    (loop for k from 0 below inlen do
	      (let ((kc (elt letters k)))
		(if (or (char-equal #\( kc)
			(char-equal #\) kc))
		    (setf (elt newstr (incf i)) #\\))
		(setf (elt newstr (incf i)) kc)))
	    newstr)
	letters))))

(defmethod show ((server server) (text text) &key (dx 0) (dy 0) font size scaler (save :both) &allow-other-keys)
  (when (not (invisible-matrix-p text))
    (let* ((lfont (or font (font-name text)))
	   (ssize (or *cmn-score-size* (scr-size server)))
	   (lsize (or size (font-size text) (* (or scaler (font-scaler text)) ssize)))
	   (pstart "")
	   (zxy (and (zerop dx) (zerop dy)))
	   (pend (if zxy "show" "ashow")))

      (when (output server)
	(let ((c-out (c-output-in-progress server)))
	  (if lfont
	      (if (< lsize 1.0)
		  (if (and (font-size text)
			   (< (font-size text) 1.0))
		      (warn "(text ~S: font-size = ~1,3F which will be invisible)" 
			    (letters text) (font-size text))
		    (warn "(text ~S: font-scaler = ~1,3F so font-size = ~1,3F which will be invisible)" 
			  (letters text) (font-scaler text) lsize))))
	  (if (and c-out zxy lfont (not (matrix text)) (not (pattern text)) (eq save :both))
	      ;; optimize most common case
	      (c-text lfont (round lsize) (ps-letters text))
	    (progn
	      (when (or lfont (matrix text)) 
		(if (member save '(:both :start)) 
		    (if c-out 
			(c-print " gsave") 
		      (format (output server) " gsave~%")))
		(if lfont 
		    (if c-out
			(c-print (format nil " /~A findfont ~D scalefont setfont" lfont lsize))
		      (format (output server) " /~A findfont ~D scalefont setfont~%" lfont lsize)))
		(if (matrix text)
		    (if c-out
			(c-print (format nil "[ ~{~A ~}] concat" (map 'list #'not-rational (matrix text))))
		      (format (output server) "[ ~{~A ~}] concat~%" (map 'list #'not-rational (matrix text))))))
	      (if (pattern text)
		  (multiple-value-bind
		      (qpstart qpend) 
		      (funcall (pattern text) server text)
		    (if qpstart (setf pstart qpstart))
		    (if qpend (setf pend qpend))))
	      (if (or zxy
		      (and pstart (not (string-equal pend "ashow"))))
		  (if c-out
		      (c-print (format nil " ~A(~A) ~A" pstart (ps-letters text) pend))
		    (format (output server) " ~A(~A) ~A~%" pstart (ps-letters text) pend))
		(if c-out
		    (c-print (format nil " ~A ~1,3F ~1,3F (~A) ashow" pstart (* dx ssize) (* dy ssize) (ps-letters text)))
		  (format (output server) " ~A ~1,3F ~1,3F (~A) ashow~%" pstart (* dx ssize) (* dy ssize) (ps-letters text))))
	      (if (and (member save '(:both :finish)) 
		       (or lfont (matrix text))) 
		  (if c-out
		      (c-print " grestore")
		    (format (output server) " grestore~%")))))))
      (%dxy server (%rx server text :dx dx :size lsize :scaler scaler) lsize)
      (%update-bounds server))))


(defmethod %rx ((server server) (text text) &key (dx 0) size scaler &allow-other-keys)
  (if (identity-matrix-p text)
      (* (length (letters text)) (+ .4 dx) (or scaler 1.0) (or size *cmn-score-size* (scr-size server)))
    0))


(defmethod curveto ((server server) x0 y0 x1 y1 x2 y2)
  (declare (optimize (speed 3) (safety 0)))
  (let* ((size (or *cmn-score-size* (scr-size server)))
	 (size-x2 (* size x2))
	 (size-y2 (* size y2)))
    (if (output server)
	(if (c-output-in-progress server)
	    (progn
	      (setf (aref c-args 0) (float (* size x0)))
	      (setf (aref c-args 1) (float (* size y0)))
	      (setf (aref c-args 2) (float (* size x1)))
	      (setf (aref c-args 3) (float (* size y1)))
	      (setf (aref c-args 4) (float size-x2))
	      (setf (aref c-args 5) (float size-y2))
	      (c-curveto c-args))
	  (format (output server) " ~1,2F ~1,2F ~1,2F ~1,2F ~1,2F ~1,2F curveto~%" 
		  (float (* size x0))
		  (float (* size y0))
		  (float (* size x1))
		  (float (* size y1))
		  (float size-x2)
		  (float size-y2))))
    (setf (scr-x server) size-x2)
    (setf (scr-y server) size-y2)
    (setf (scr-x0 server) (min (scr-x0 server) size-x2))
    (setf (scr-y0 server) (min (scr-y0 server) size-y2))
    (setf (scr-x1 server) (max (scr-x1 server) size-x2))
    (setf (scr-y1 server) (max (scr-y1 server) size-y2))))


#|
;;; here is the PostScript Bezier curve (curveto above) done by hand:
;;; (n is how accurate a curve we draw)

(defun bezier (score x0 y0 x1 y1 x2 y2 x3 y3 n)
  (let* ((cx (* 3 (- x1 x0)))
	 (cy (* 3 (- y1 y0)))
	 (bx (- (* 3 (- x2 x1)) cx))
	 (by (- (* 3 (- y2 y1)) cy))
	 (ax (- x3 x0 cx bx))
	 (ay (- y3 y0 cy by))
	 (incr (/ 1.0 n)))
    (moveto score x0 y0)
    (loop for i from 0 to 1 by incr do
      (lineto score (+ x0 (* i (+ cx (* i (+ bx (* i ax))))))
	      (+ y0 (* i (+ cy (* i (+ by (* i ay))))))))
    (draw score)))

|#

(defun solve-bezier (x x0 y0 y1 y2 x3 y3)
  (let* ((cy (* 3 (- y1 y0)))
	 (by (- (* 3 (- y2 y1)) cy))
	 (ay (- y3 y0 cy by))
	 (i (divide (- x x0) (- x3 x0))))
    (+ y0 (* i (+ cy (* i (+ by (* i ay))))))))

(defmethod fill-in ((server server) &key rectangle closepath path)
  (if (output server)
      (let ((cmd (if rectangle (if (= (length path) 4) 
				   "rectfill" 
				 (error "internal cmn error -- rectangle of ~D corners?" (length path)))
		   (if closepath "closepath fill"
		     "fill")))
	    (c-out (c-output-in-progress server)))
	(if path
	    (let ((size (or *cmn-score-size* (scr-size server))))
	      (if c-out
		  (if (= (length path) 4)
		      (progn
			(setf (aref c-args 0) (float (* size (first path))))
			(setf (aref c-args 1) (float (* size (second path))))
			(setf (aref c-args 2) (float (* size (third path))))
			(setf (aref c-args 3) (float (* size (fourth path))))
			(c-fill-in c-args cmd))
		    (c-print (format nil " ~{~1,2F ~1,2F lineto ~}~A" (map 'list #'(lambda (n) (float (* n size))) path) cmd)))
		(if rectangle
		    (format (output server) " ~1,2F ~1,2F ~1,2F ~1,2F rectfill~%"
			    (float (* size (first path))) (float (* size (second path)))
			    (float (* size (third path))) (float (* size (fourth path))))
		  (format (output server) " ~{~1,2F ~1,2F lineto ~}~A~%" 
			  (map 'list #'(lambda (n) (float (* n size))) path)
			  cmd))))
	  (if c-out
	      (c-print cmd)
	    (format (output server) " ~A~%" cmd))))))


(defmethod circle ((score server) x y r &optional (ang1 0) (ang2 360))
  (let* ((size (scr-size score)))
    (if (output score)
	(if (c-output-in-progress score)
	    (c-print (format nil " ~1,3F ~1,3F ~D ~D ~D newpath arc" (float (* size x)) (float (* size y))
			     (float (* size r))	ang1 ang2))
	  (format (output score) " ~1,3F ~1,3F ~D ~D ~D newpath arc~%" 
		  (float (* size x))
		  (float (* size y))
		  (float (* size r))
		  ang1 ang2)))
    (setf (bounds score) (list (* (- x r) size)
			       (* (- y r) size)
			       (* (+ x r) size)
			       (* (+ y r) size)))))

(defmethod optimize-stem ((score server) x0 y0 y1 width)
  (let* ((ssize (scr-size score))
	 (x0s (float (* ssize x0)))
	 (y0s (float (* ssize y0)))
	 (y1s (float (* ssize y1)))
	 (sws (float (* ssize width))))
    (if (output score)
	(if (c-output-in-progress score)
	    (progn
	      (setf (aref c-args 0) x0s)
	      (setf (aref c-args 1) y0s)
	      (setf (aref c-args 2) y1s)
	      (setf (aref c-args 3) sws)
	      (c-stem c-args))
	  (format (output score) " ~1,3F setlinewidth ~1,2F ~1,2F moveto ~1,2F ~1,2F lineto stroke 0 setlinewidth~%"
		  sws x0s y0s x0s y1s)))
    (%xy score x0s y1s)
    (%update-bounds score)))

(defmethod just-stem ((score server) x0 y0 y1)
  (let* ((ssize (scr-size score))
	 (x0s (float (* ssize x0)))
	 (y0s (float (* ssize y0)))
	 (y1s (float (* ssize y1))))
    (if (output score)
	(if (c-output-in-progress score)
	    (progn
	      (setf (aref c-args 0) x0s)
	      (setf (aref c-args 1) y0s)
	      (setf (aref c-args 2) y1s)
	      (c-just-stem c-args))
	  (format (output score) " ~1,2F ~1,2F moveto ~1,2F ~1,2F lineto~%" x0s y0s x0s y1s)))))

(defmethod just-line ((score server) x0 y0 x1)
  (let* ((ssize (scr-size score))
	 (x0s (float (* ssize x0)))
	 (y0s (float (* ssize y0)))
	 (x1s (float (* ssize x1))))
    (if (output score)
	(if (c-output-in-progress score)
	    (progn
	      (setf (aref c-args 0) x0s)
	      (setf (aref c-args 1) y0s)
	      (setf (aref c-args 2) x1s)
	      (c-line c-args))
	  (format (output score) " ~1,2F ~1,2F moveto ~1,2F ~1,2F lineto~%" x0s y0s x1s y0s)))))

(defmethod slanted-fill-in ((score server) x0 y0 x1 y1 thickness)
  (let* ((ssize (scr-size score))
	 (x0s (float (* ssize x0)))
	 (y0s (float (* ssize y0)))
	 (x1s (float (* ssize x1)))
	 (y1s (float (* ssize y1)))
	 (thk (float (* ssize thickness))))
    (if (output score)
	(if (c-output-in-progress score)
	    (progn
	      (setf (aref c-args 0) x0s)
	      (setf (aref c-args 1) y0s)
	      (setf (aref c-args 2) x1s)
	      (setf (aref c-args 3) y1s)
	      (setf (aref c-args 4) thk)
	      (c-slanted-line c-args))
	  (format (output score) " ~1,2F ~1,2F moveto ~1,2F ~1,2F lineto ~1,2F ~1,2F lineto ~1,2F ~1,2F lineto closepath fill~%"
		  x0s y0s x0s (+ y0s thk) x1s (+ y1s thk) x1s y1s)))
    (setf (scr-x score) x0s)
    (setf (scr-y score) y0s)
    (setf (scr-x0 score) (min (scr-x0 score) x0s x1s))
    (setf (scr-y0 score) (min (scr-y0 score) y0s y1s))
    (setf (scr-x1 score) (max (scr-x1 score) x0s x1s))
    (setf (scr-y1 score) (max (scr-y1 score) (+ (max y0s y1s) thk)))))




(defun month-name (month) (nth (- month 1) '("Jan" "Feb" "Mar" "Apr" "May" "Jun" "Jul" "Aug" "Sep" "Oct" "Nov" "Dec")))
(defun day-name (day) (nth day '("Mon" "Tue" "Wed" "Thu" "Fri" "Sat" "Sun")))

(defun creation-date ()
  (multiple-value-bind (second minute hour date month year day daylight-saving-p time-zone)
      (get-decoded-time)
    (declare (ignore second daylight-saving-p time-zone))
    (format nil "~A ~D-~A-~D at ~D:~2,'0D"
	    (day-name day) date (month-name month) (- year 1900) hour minute)))

(defmethod header ((server server) &key name &allow-other-keys)
  (if (not (c-output-in-progress server))
      (progn
	(format (output server) "%!PS-Adobe-2.0 EPSF-2.0~%") ;this is what our Next's seem to have -- Adobe recommends 3.0
	;; exact caps matter here --if you had %!Ps-Adobe, WriteNow would throw up its hands and refuse to paste it in.
	(if name (format (output server) "%%Title: ~A~%" name))
	#+(or Excl KCL)	(format (output server) "%%Creator: ~A~%" (first (last (pathname-directory (user-homedir-pathname)))))
	(format (output server) "%%CreationDate: ~A~%" (creation-date))
	(if (bounded server)
	    (progn
	      (setf (bounds server) (bounded server))
	      (format (output server) "%%BoundingBox:~{ ~D~}~%" (map 'list #'ceiling (bounds server))))
	  (format (output server) "%%BoundingBox:(atend)~%"))
	(format (output server) "%%EndComments~%")
	(include-eps-defs-and-whatever server)
	(format (output server) "%%EndProlog~%"))
    (progn
      (c-print (format nil "%!PS-Adobe-2.0 EPSF-2.0"))
      (if name (c-print (format nil "%%Title: ~A" name)))
      (c-print (format nil "%%Creator: ~A" (caddr (pathname-directory (user-homedir-pathname)))))
      (c-print (format nil "%%CreationDate: ~A" (creation-date)))
      (if (bounded server)
	  (progn
	    (setf (bounds server) (bounded server))
	    (c-print (format nil "%%BoundingBox:~{ ~D~}" (map 'list #'ceiling (bounds server)))))
	(c-print (format nil "%%BoundingBox:(atend)")))
      (c-print (format nil "%%EndComments"))
      (include-eps-defs-and-whatever server)
      (c-print (format nil "%%EndProlog")))))


(defmethod footer ((server server) &key)
  (if (c-output-in-progress server)
      (c-print (format nil " showpage~%%%Trailer"))
    (format (output server) " showpage~%%%Trailer~%"))
					;  (incf (box-y1 server) 10)
  (if (not (bounded server))
      (progn
	;; should we pad x1 out to the page-width here? what about header-margin?
	(setf (box-y1 server) (max (box-y1 server)
				   (min (+ (box-y1 server) (inches-to-ps (header-margin *cmn-score*)))
					(inches-to-ps (page-height *cmn-score*)))))
	(setf (box-y0 server) (min (box-y0 server)
				   (max 0 (- (box-y0 server) (inches-to-ps (footer-margin *cmn-score*))))))
	(setf (box-x0 server) 0)
	(if (and (lines server) (> (lines server) 1))
	    (setf (box-x1 server) (max (box-x1 server) ;for /NextApps/Preview's benefit
				       (inches-to-ps (page-width *cmn-score*))))
	  (if (< (box-x1 server) (inches-to-ps (- (page-width *cmn-score*) (right-margin *cmn-score*))))
	      (incf (box-x1 server) (inches-to-ps (right-margin *cmn-score*)))))
	(if (c-output-in-progress server)
	    (c-print (format nil "%%BoundingBox:~{ ~D~}~%" (map 'list #'(lambda (n) 
									  (round (1+ n)))
								(bounds server))))
	  (format (output server) "%%BoundingBox:~{ ~D~}~%" (map 'list #'(lambda (n) 
									   (round (1+ n)))
								 (bounds server)))))))

(defmethod initialize ((server server) &key (file "aaa.eps") font size title bounded with-pages)
  (if (not (output-file server)) (setf (output-file server) file))
  (if (not (pathname-type (output-file server))) (setf (output-file server) (concatenate 'string (output-file server) ".eps")))
  (if (c-output-in-progress server)
      (let ((ok 0))
	(setf (output server) t)
	(setf ok (c-open (output-file server)))
	(if (= ok 1) (throw 'cannot-open-file nil)))
    (setf (output server) (open (output-file server)
				:direction :output
				:if-exists :supersede
				:if-does-not-exist :create)))
  (setf (bounded server) bounded)
  (if size (setf (scr-size server) size)
    (if (not (scr-size server)) (setf (scr-size server) 40)))
  (if (not *cmn-score-size*) (setf *cmn-score-size* (scr-size server)))
  (if font (setf (font server) font)
    (if (not (font server)) (setf (font server) Music-Font)))
  (header server :name title)
  (if (c-output-in-progress server)
      (progn
	(if with-pages (c-print (format nil "%%Page: 1 1")))
	(c-print (format nil " /~A findfont ~D scalefont setfont" (font server) (scr-size server)))
	(c-print (format nil " 0 setlinewidth")))
    (progn
      (if with-pages (format (output server) "%%Page: 1 1~%"))
      (format (output server) " /~A findfont ~D scalefont setfont~%" (font server) (scr-size server))
      (format (output server) " 0 setlinewidth~%")))
  server)


(defmethod clear-server ((server server))
  (declare (optimize (speed 3) (safety 0)))
  (setf (output server) nil)
  (clear-bounding-box server)
  (setf (scr-x server) 0)
  (setf (scr-y server) 0))

(defmethod finalize ((server server) &key)
  (footer server)
  (if (c-output-in-progress server)
      (progn
	(c-close)
	(setf (c-output-in-progress server) nil))
    (close (output server)))
  (setf (output-file server) nil)
  (clear-server server)
  (setf *cmn-score-size* nil))



;;;
;;; --------------    Font info
;;; 

;;; now infinite data about Sonata font -- all in Sonata.afm (or whatever).
;;; The first constants below are OCTAL mascarading as decimal -- my mistake.
;;; Adobe font metric info = octal char code, rx dx dy w h
;;;
;;; a few of these numbers have been hand-tweaked (%staff for example) -- 
;;; don't read directly from the .afm file!


(defun make-constant-glyph (info)
  (if info
      (make-instance 'write-protected-glyph 
	  :index (first info) 
	  :rx (* .001 (second info))
	  :x0 (* .001 (third info))
	  :y0 (* .001 (fourth info))
	  :x1 (* .001 (fifth info))
	  :y1 (* .001 (sixth info)))
    (warn "nil argument to make-constant-glyph")))

(defparameter %zero (make-constant-glyph #+Sonata '(60 360 0 -217 360 218) #+Petrucci-Medium '(60 244 1 -253 230 253)))
(defparameter %one (make-constant-glyph #+Sonata '(61 250 0 -209 250 209) #+Petrucci-Medium '(61 378 0 -253 364 253)))
(defparameter %two (make-constant-glyph #+Sonata '(62 340 0 -217 340 218) #+Petrucci-Medium '(62 330 0 -253 316 253)))
(defparameter %three (make-constant-glyph #+Sonata '(63 324 0 -217 324 218) #+Petrucci-Medium '(63 348 0 -253 335 253)))
(defparameter %four (make-constant-glyph #+Sonata '(64 329 0 -209 329 209) #+Petrucci-Medium '(64 316 2 -253 303 253)))
(defparameter %five (make-constant-glyph #+Sonata '(65 300 0 -209 300 209) #+Petrucci-Medium '(65 344 2 -253 330 253)))
(defparameter %six (make-constant-glyph #+Sonata '(66 334 0 -217 334 218) #+Petrucci-Medium '(66 347 4 -253 335 253)))
(defparameter %seven (make-constant-glyph #+Sonata '(67 341 0 -215 341 216) #+Petrucci-Medium '(67 327 2 -253 314 253)))
(defparameter %eight (make-constant-glyph #+Sonata '(70 328 0 -217 328 218) #+Petrucci-Medium '(70 342 0 -253 328 253)))
(defparameter %nine (make-constant-glyph #+Sonata '(71 334 0 -217 334 218) #+Petrucci-Medium '(71 287 10 -445 165 96)))

(defparameter %sharp (make-constant-glyph #+Sonata '(43 224 0 -379 224 375) #+Petrucci-Medium '(43 237 1 -380 211 380)))
(defparameter %smallsharp (make-constant-glyph #+Sonata '(111 112 0 -193 112 188) #+Petrucci-Medium '(111 215 30 -185 136 214)))
(defparameter %flat (make-constant-glyph #+Sonata '(142 198 0 -178 198 491) #+Petrucci-Medium '(142 224 2 -195 219 485)))
(defparameter %smallflat (make-constant-glyph #+Sonata '(151 95 0 -89 95 246) #+Petrucci-Medium '(151 224 12 -123 122 216)))
(defparameter %natural (make-constant-glyph #+Sonata '(156 176 0 -350 176 350) #+Petrucci-Medium '(156 195 1 -349 170 349)))
(defparameter %dblflat (make-constant-glyph #+Sonata '(272 367 0 -178 367 491) #+Petrucci-Medium '(272 403 2 -195 402 485)))
(defparameter %dblsharp (make-constant-glyph #+Sonata '(334 266 0 -131 266 131) #+Petrucci-Medium '(334 278 0 -126 253 126)))
(defparameter %smallnatural (make-constant-glyph #+Sonata '(351 88 0 -192 88 186) #+Petrucci-Medium '(351 169 0 -169 84 177)))

(defparameter %pedal (make-constant-glyph #+Sonata '(241 792 0 -2 792 458) #+Petrucci-Medium '(241 824 0 -3 808 507)))
(defparameter %pedaloff (make-constant-glyph #+Sonata '(52 361 0 0 361 350) #+Petrucci-Medium '(52 406 -4 4 374 376)))
(defparameter %mordentslash (make-constant-glyph #+Sonata '(115 586 0 -90 586 310) #+Petrucci-Medium '(115 634 0 -84 607 338)))
(defparameter %harmonic (make-constant-glyph #+Sonata '(117 300 0 -131 300 132) #+Petrucci-Medium '(117 301 0 -126 269 127)))
(defparameter %turn (make-constant-glyph #+Sonata '(124 618 0 -133 618 133) #+Petrucci-Medium '(124 621 -25 -167 605 130)))
(defparameter %arpeggio (make-constant-glyph #+Sonata '(147 138 0 0 138 534) #+Petrucci-Medium '(147 152 9 5 136 571)))
(defparameter %mordent (make-constant-glyph #+Sonata '(155 606 0 -1 606 240) #+Petrucci-Medium '(155 528 0 7 607 257)))
(defparameter %naturalharmonic (make-constant-glyph #+Sonata '(157 120 0 -61 120 60) #+Petrucci-Medium '(157 203 0 -84 169 84)))
(defparameter %fermataup (make-constant-glyph #+Sonata '(165 658 0 -352 658 0) #+Petrucci-Medium '(165 710 -2 -383 678 -2)))
(defparameter %trillsection (make-constant-glyph #+Sonata '(176 337 0 0 413 165) #+Petrucci-Medium '(176 338 -45 -7 383 173)))
(defparameter %mordentthree (make-constant-glyph #+Sonata '(265 852 0 -1 852 238) #+Petrucci-Medium '(265 790 0 5 870 257)))
(defparameter %tremeloshort (make-constant-glyph #+Sonata '(276 300 0 0 300 479) #+Petrucci-Medium '(276 379 1 0 335 497)))
(defparameter %trillofinno (make-constant-glyph #+Sonata '(331 468 0 -13 468 369) #+Petrucci-Medium '(331 444 0 -22 423 380)))

(defparameter %coda (make-constant-glyph #+Sonata '(336 443 0 -254 443 255) #+Petrucci-Medium '(336 482 19 -66 442 429)))
(defparameter %repeatdots (make-constant-glyph #+Sonata '(173 126 0 307 126 696) #+Petrucci-Medium '(40 150 0 0 0 0)))
(defparameter %segno (make-constant-glyph #+Sonata '(45 515 0 -343 515 336) #+Petrucci-Medium '(45 570 12 0 532 694)))
(defparameter %repsign (make-constant-glyph #+Sonata '(324 610 0 0 610 509) #+Petrucci-Medium '(324 676 0 0 633 507)))

(defparameter %trebleclef (make-constant-glyph #+Sonata '(46 611 0 -402 611 1334) #+Petrucci-Medium '(46 676 0 -684 670 1077)))
(defparameter %treble8clef (make-constant-glyph #+Sonata '(46 611 0 -402 611 1334) #+Petrucci-Medium '(126 687 12 -895 683 1077)))
(defparameter %percclef1 (make-constant-glyph #+Sonata '(57 174 0 0 174 500) #+Petrucci-Medium '(57 190 0 0 169 507)))
(defparameter %bassclef (make-constant-glyph #+Sonata '(77 691 0 192 691 991) #+Petrucci-Medium '(77 723 4 -525 712 248)))
(defparameter %cclef (make-constant-glyph #+Sonata '(102 674 0 -9 674 1009) #+Petrucci-Medium '(102 676 0 -507 694 507)))

(defparameter %space (make-constant-glyph #+Sonata '(40 150 0 0 0 0) #+Petrucci-Medium '(40 150 0 0 0 0)))
(defparameter %parenleft (make-constant-glyph #+Sonata '(50 165 0 -50 165 500) #+Petrucci-Medium '(50 241 0 21 190 570)))
(defparameter %parenright (make-constant-glyph #+Sonata '(51 165 0 -50 165 500) #+Petrucci-Medium '(51 241 5 21 195 570)))
(defparameter %cross (make-constant-glyph #+Sonata '(53 300 0 -150 300 150) #+Petrucci-Medium '(53 330 -42 -186 338 189)))

(defparameter %fermata (make-constant-glyph #+Sonata '(125 658 0 0 658 352) #+Petrucci-Medium '(125 697 -2 0 678 380)))
(defparameter %grandpause (make-constant-glyph #+Sonata '(42 511 0 0 511 509) #+Petrucci-Medium '(42 494 14 0 450 507)))
(defparameter %pause (make-constant-glyph #+Sonata '(54 182 0 -231 182 68) #+Petrucci-Medium '(54 211 23 -218 185 53)))

(defparameter %wedgefill (make-constant-glyph #+Sonata '(340 228 0 0 228 199) #+Petrucci-Medium '(340 330 0 0 295 211)))
(defparameter %dot (make-constant-glyph #+Sonata '(56 94 0 -46 94 48) #+Petrucci-Medium '(56 104 22 -16 101 59)))
(defparameter %accent (make-constant-glyph #+Sonata '(76 417 0 4 417 257) #+Petrucci-Medium '(76 418 0 0 390 169)))
(defparameter %dotaccentup (make-constant-glyph #+Sonata '(254 278 0 0 278 416) #+Petrucci-Medium '(254 317 0 0 295 422)))
(defparameter %dotaccentdown (make-constant-glyph #+Sonata '(350 278 0 0 278 416) #+Petrucci-Medium '(350 317 0 11 295 433)))
(defparameter %wedgedown (make-constant-glyph #+Sonata '(166 263 0 0 263 417) #+Petrucci-Medium '(166 304 -5 -8 290 414)))
(defparameter %wedgeup (make-constant-glyph #+Sonata '(136 263 0 0 263 417) #+Petrucci-Medium '(136 317 0 0 295 422)))

(defparameter %eighthflagup (make-constant-glyph #+Sonata '(152 0 275 124 546 874) #+Petrucci-Medium '(152 331 0 -543 304 269)))
(defparameter %eighthflagdown (make-constant-glyph #+Sonata '(112 0 0 -874 271 -124) #+Petrucci-Medium '(112 331 0 -269 304 543)))
(defparameter %dblwholenote (make-constant-glyph #+Sonata '(127 614 0 -166 614 166) #+Petrucci-Medium '(127 637 7 -183 636 183)))
(defparameter %wholenote (make-constant-glyph #+Sonata '(167 412 0 -126 412 126) #+Petrucci-Medium '(167 406 0 -126 405 126)))
(defparameter %quarternotehead (make-constant-glyph #+Sonata '(317 300 0 -125 300 125) #+Petrucci-Medium '(317 287 1 -131 283 124)))
(defparameter %extendflagdown (make-constant-glyph #+Sonata '(360 0 0 -159 272 375) #+Petrucci-Medium '(360 312 0 -249 296 466)))
(defparameter %halfnotehead (make-constant-glyph #+Sonata '(372 300 0 -134 300 134) #+Petrucci-Medium '(372 283 1 -131 283 124)))
(defparameter %extendflagup (make-constant-glyph #+Sonata '(373 0 275 -360 546 184) #+Petrucci-Medium '(373 315 0 -466 296 249)))
(defparameter %rhythmX (make-constant-glyph #+Sonata '(300 300 0 -149 300 150) #+Petrucci-Medium '(300 292 -42 -175 301 175)))

(defparameter %staff (make-constant-glyph #+Sonata '(75 1000 -11 -9 1011 1009) #+Petrucci-Medium '(75 1014 -42 0 1056 1056)))
(defparameter %bottombracket (make-constant-glyph #+Sonata '(114 408 0 -296 408 155) #+Petrucci-Medium '(40 150 0 0 0 0)))
(defparameter %bracedown (make-constant-glyph #+Sonata '(352 300 0 -1001 300 1009) #+Petrucci-Medium '(40 150 0 0 0 0)))
(defparameter %topbracket (make-constant-glyph #+Sonata '(302 408 0 -155 408 296) #+Petrucci-Medium '(40 150 0 0 0 0)))
(defparameter %braceup (make-constant-glyph #+Sonata '(247 300 0 -1001 300 1009) #+Petrucci-Medium '(40 150 0 0 0 0)))

(defparameter %cuttime (make-constant-glyph #+Sonata '(103 421 0 -424 421 425) #+Petrucci-Medium '(103 423 4 -414 418 376)))
(defparameter %commontime (make-constant-glyph #+Sonata '(143 421 0 -254 421 255) #+Petrucci-Medium '(143 423 4 -248 418 247)))

(defparameter %mezzoforte (make-constant-glyph #+Sonata '(106 834 0 -172 834 413) #+Petrucci-Medium '(106 875 5 -177 859 419)))
(defparameter %mezzopiano (make-constant-glyph #+Sonata '(120 800 0 -172 800 266) #+Petrucci-Medium '(120 837 5 -147 805 295)))
(defparameter %sforzando (make-constant-glyph #+Sonata '(123 599 0 -172 599 413) #+Petrucci-Medium '(123 659 11 -148 614 448)))
(defparameter %forzando (make-constant-glyph #+Sonata '(132 651 0 -172 651 413) #+Petrucci-Medium '(132 685 1 -177 669 419)))
(defparameter %forte (make-constant-glyph #+Sonata '(146 515 0 -172 515 413) #+Petrucci-Medium '(146 518 1 -177 516 419)))
(defparameter %piano (make-constant-glyph #+Sonata '(160 445 0 -172 445 266) #+Petrucci-Medium '(160 437 0 -147 396 295)))
(defparameter %s (make-constant-glyph #+Sonata '(163 226 0 -16 226 266) #+Petrucci-Medium '(163 259 11 -9 236 292)))
(defparameter %z (make-constant-glyph #+Sonata '(172 291 0 -16 291 250) #+Petrucci-Medium '(172 355 -2 -12 290 287)))
(defparameter %ppp (make-constant-glyph #+Sonata '(270 1197 0 -172 1197 266) #+Petrucci-Medium '(270 1206 0 -147 1100 295)))
(defparameter %pp (make-constant-glyph #+Sonata '(271 821 0 -172 821 266) #+Petrucci-Medium '(271 823 0 -147 748 295)))
(defparameter %m (make-constant-glyph #+Sonata '(275 479 0 -16 479 266) #+Petrucci-Medium '(275 520 5 -13 471 294)))
(defparameter %ff (make-constant-glyph #+Sonata '(304 781 0 -172 781 413) #+Petrucci-Medium '(304 792 1 -177 788 419)))
(defparameter %fff (make-constant-glyph #+Sonata '(354 1047 0 -172 1047 413) #+Petrucci-Medium '(354 1063 1 -177 1060 419)))
(defparameter %r (make-constant-glyph #+Sonata '(363 280 0 0 280 266) #+Petrucci-Medium '(40 150 0 0 0 0)))

(defparameter %thirtyrest (make-constant-glyph #+Sonata '(250 207 0 -384 400 578) #+Petrucci-Medium '(250 423 0 -511 377 456)))
(defparameter %wholerest (make-constant-glyph #+Sonata '(267 313 0 125 313 250) #+Petrucci-Medium '(267 423 0 126 422 255)))
(defparameter %sixteenthrest (make-constant-glyph #+Sonata '(305 138 0 -384 331 322) #+Petrucci-Medium '(305 304 0 -511 318 202)))
(defparameter %measurerest (make-constant-glyph #+Sonata '(335 500 0 0 500 500) #+Petrucci-Medium '(335 515 0 -253 507 253)))
(defparameter %dblwholerest (make-constant-glyph #+Sonata '(343 150 0 0 150 250) #+Petrucci-Medium '(40 150 0 0 0 0)))
(defparameter %eighthrest (make-constant-glyph #+Sonata '(344 69 0 -384 262 78) #+Petrucci-Medium '(344 304 28 -257 287 202)))
(defparameter %one28rest (make-constant-glyph #+Sonata '(345 345 0 -384 538 1078) #+Petrucci-Medium '(345 608 0 -511 495 957)))
(defparameter %halfrest (make-constant-glyph #+Sonata '(356 313 0 0 313 125) #+Petrucci-Medium '(356 423 0 -2 422 126)))
(defparameter %sixtyfourrest (make-constant-glyph #+Sonata '(364 276 0 -384 469 828) #+Petrucci-Medium '(364 608 0 -511 436 709)))
(defparameter %quarterrest (make-constant-glyph #+Sonata '(316 257 0 -355 257 400) #+Petrucci-Medium '(316 313 12 -404 290 388)))

(defparameter %square (make-constant-glyph #+Sonata '(255 300 0 -150 300 150) #+Petrucci-Medium '(255 279 -42 -169 295 169)))
(defparameter %squarefill (make-constant-glyph #+Sonata '(320 300 0 -150 300 150) #+Petrucci-Medium '(320 308 0 -118 278 160)))
(defparameter %trianglefill (make-constant-glyph #+Sonata '(321 300 0 -116 300 155) #+Petrucci-Medium '(321 389 0 -135 371 143)))
(defparameter %diamond (make-constant-glyph #+Sonata '(341 300 0 -150 300 150) #+Petrucci-Medium '(341 275 -42 -173 304 173)))
(defparameter %diamondfill (make-constant-glyph #+Sonata '(342 300 0 -150 300 150) #+Petrucci-Medium '(342 299 0 -142 283 142)))



(defun digit-to-glyph (num)
  (case num (0 %zero) (1 %one) (2 %two) (3 %three) (4 %four) (5 %five) (6 %six) (7 %seven) (8 %eight) (9 %nine)))

(defun number-to-glyph (num)
  (if (<= 0 num 9) 
      (digit-to-glyph num)
    (text (format nil "~D" num))))	;font happens to line up right -- probably not the "right" thing in general.

(defun character-to-glyph (char)
  (case char
    (#\0 %zero) (#\1 %one) (#\2 %two) (#\3 %three) (#\4 %four)
    (#\5 %five) (#\6 %six) (#\7 %seven) (#\8 %eight) (#\9 %nine)
    (#\+ %cross) (#\. %dot) (#\C %commontime) (#\f %forte)
    (#\s %s) (#\r %r) (#\z %z) (#\p %piano) (#\m %m)
    (otherwise %space)))




;;; as a convenience for callers of clm, we like to package up all the information about
;;; an entity within the parentheses of that entity -- that is:
;;;    (c4 q staccato (onset 1.5))
;;; which requires that the "onset" function pass its argument on to the "c4" function.
;;; In normal CLOS syntax this would be (setf (onset (c4 q staccato)) 1.5) or something,
;;; and the real entity rapidly gets swallowed up in onion-like layers of setfs with the
;;; associated setf-d data further and further from the field it is setting.  The
;;; functions that act like "onset" above return a "self-acting" instance and the
;;; "self-action" macro wraps up the rest.

(defstruct (self-acting (:conc-name nil)) action arguments)

(eval-when (compile load eval)
  (defmacro self-action (name setf-name)
    `(progn
       (defgeneric ,name (obj))
       (defgeneric ,setf-name (obj val))
       (defgeneric (setf ,name) (val obj))
       (defmethod ,setf-name ((inst t) val) (setf (,name inst) val))
       (defmethod ,name ((val t)) (make-self-acting :action #',setf-name :arguments val))))
  )

(defun rotate (object &optional angle)
  (if (or angle
	  (not (numberp object)))
      (%rotate object angle)
    (make-self-acting :action #'%rotate :arguments object)))

(defun scale (object &optional xscl yscl)
  (if (or yscl
	  (not (numberp object)))
      (%scale object xscl yscl)
    (make-self-acting :action #'%scale-1 :arguments (list object xscl))))

(defun mirror (&optional object)
  (if (or object
	  (not (numberp object)))
      (%mirror object)
    (make-self-acting :action #'%mirror :arguments nil)))

(defvar mirrored (make-self-acting :action #'%mirror :arguments nil))

(defun transform (object &optional matrix)
  (if (or matrix
	  (not (listp object)))
      (%transform object matrix)
    (make-self-acting :action #'%transform :arguments object)))




(defclass bounds-mixin ()		;used for alignment and justification
  ((center :initarg :center		;center point
	   :initform 0
	   :reader center)
   (walls :initarg :walls		;absolute "don't draw inside these bounds" 
	  :initform nil
	  :reader walls)
   (fences :initarg :fences		;ideal bounds if there is room for them
	   :initform nil
	   :reader fences)
   (expanders :initform nil		;expansion behaviour when there's lots of free room to parcel out
	      :initarg :expanders
	      :reader expanders)))

(defclass bounds (bounds-mixin)
  ((center :accessor center)
   (walls :accessor walls)
   (fences :accessor fences)
   (expanders :accessor expanders)))

(defmethod descry ((bounds bounds-mixin) &optional stream controller)
  (format stream "~A~A~A~A~A~A~A"
	  (if (not controller) "(bounds" 
	    (if (or (walls bounds) (fences bounds) (expanders bounds)) 
		(format nil "~%~A" prewhitespace)
	      ""))
	  (if (and (center bounds) (not (zerop (center bounds)))) (format nil " :center ~A" (center bounds)) "")
	  (if (walls bounds) (format nil " :walls '~A" (walls bounds)) "")
	  (if (fences bounds) (format nil " :fences '~A" (fences bounds)) "")
	  (if (expanders bounds) (format nil " :expanders '~A" (expanders bounds)) "")
	  (if (next-method-p) (call-next-method bounds stream (or controller bounds)) "")
	  (if (not controller) ")" "")))

(defmethod copy ((bounds bounds-mixin) &optional object)
  (let ((new-object (if (not object) (make-bounds)
		      (if (write-protected object) 
			  (copy object) 
			object))))
    (setf (center new-object) (center bounds))
    (setf (walls new-object) (copy-list (walls bounds)))
    (setf (fences new-object) (copy-list (fences bounds)))
    (setf (expanders new-object) (copy-list (expanders bounds)))
    (if (next-method-p) (call-next-method bounds new-object))
    new-object))

(self-action center setf-center)
(self-action fences setf-fences)  
(self-action walls setf-walls)
(self-action expanders setf-expanders)  

(self-action x0 setf-x0) 
(self-action y0 setf-y0) 
(self-action x1 setf-x1) 
(self-action y1 setf-y1) 

(self-action dx setf-dx)
(self-action dy setf-dy)

(self-action staff-y0 setf-staff-y0)

(self-action matrix setf-matrix) 
(self-action pattern setf-pattern) 

(self-action font-name setf-font-name) 
(self-action font-size setf-font-size) 
(self-action font-scaler setf-font-scaler) 



(defclass odb-mixin ()
  ((onset :initarg :onset :initform nil :reader onset :reader odb-onset)
   (duration :initarg :duration :initform nil :reader duration :reader odb-duration)
   (beat :initarg :beat :initform nil :reader beat :reader odb-beat)))

(defclass odb (odb-mixin)
  ((onset :accessor onset :accessor odb-onset)
   (duration :accessor duration :accessor odb-duration)
   (beat :accessor beat :accessor odb-beat)))

(defmethod descry ((odb odb-mixin) &optional stream controller)
  (format stream "~A~A~A~A~A~A"
	  (if (not controller) "(odb" 
	    (if (or (onset odb) (duration odb) (beat odb))
		(format nil "~%~A" prewhitespace)
	      ""))
	  (if (onset odb) (format nil " :onset ~A" (onset odb)) "")
	  (if (duration odb) (format nil " :duration ~A" (duration odb)) "")
	  (if (beat odb) (format nil " :beat ~A" (beat odb)) "")
	  (if (next-method-p) (call-next-method odb stream (or controller odb)) "")
	  (if (not controller) ")" "")))

(defmethod copy ((odb odb-mixin) &optional object)
  (let ((new-object (if (not object) (make-odb) 
		      (if (write-protected object) (copy object)
			object))))
    (setf (odb-onset new-object) (odb-onset odb))
    (setf (odb-duration new-object) (odb-duration odb))
    (setf (odb-beat new-object) (odb-beat odb))
    (if (next-method-p) (call-next-method odb new-object))
    new-object))

;;;(self-action onset setf-onset)
;;;(self-action duration setf-duration)
;;;(self-action beat setf-beat)
;;; these three have to be careful -- avoid later round-off errors

#-KCL
(defun ratify (num)			;rational returns gigantic useless factors
  (if (floatp num)
      (loop for i from 1 to 64 and val from num by num do 
	(if (< (abs (- val (round val))) .015625) ;256-th note (1/64)
	    (return-from ratify (list (round val) i))))
    (if (ratiop num)
	(list (numerator num) (denominator num))
      (list num 1))))

#+KCL
(defun ratify (num)			;uh oh... KCL gets confused about intermediate variable types in the loop above
  (if (floatp num)
      (do ((i 1 (1+ i))
	   (val num (+ val num)))
	  ((< (abs (- val (round val))) .015625)
	   (list (round val) i))
	())
    (if (ratiop num)
	(list (numerator num) (denominator num))
      (list num 1))))



(defun fratify (num)
  (if (floatp num)
      (apply #'/ (ratify num))
    num))

;;; expand self-action by hand and make these guys use ratify
(defgeneric onset (val) )
(defgeneric duration (val) )
(defgeneric beat (val) )
(defgeneric setf-onset (obj val) )
(defgeneric setf-duration (obj val) )
(defgeneric setf-beat (obj val) )
(defgeneric (setf onset) (val obj) )
(defgeneric (setf duration) (val obj) )
(defgeneric (setf beat) (val obj) )

(defmethod onset ((val number)) (make-self-acting :action #'setf-onset :arguments val))
(defmethod duration ((val number)) (make-self-acting :action #'setf-duration :arguments val))
(defmethod beat ((val number)) (make-self-acting :action #'setf-beat :arguments val))
(defmethod setf-onset (obj val) (setf (onset obj) (fratify val)))
(defmethod setf-duration (obj val) (setf (duration obj) (fratify val)))
(defmethod setf-beat (obj val) (setf (beat obj) (fratify val)))

(defun prolog (func) (make-self-acting :action #'setf-prolog :arguments func))
(defun setf-prolog (obj val) (push val (prologue obj)))


;;;
;;; ----------------    score (top node of entire data structure)
;;;

(defclass score (server dxy bounds visible)
  ((systems :initarg :systems :accessor systems :initform nil)
   (title :initarg :title :accessor title :initform nil)
   (time-line :initarg :time-line :initform nil :accessor time-line)
   (time-lines :initarg :time-lines :initform nil :accessor time-lines)
   (line-data :initarg :line-data :initform nil :accessor line-data)
   (lines :initarg :lines :initform nil :accessor lines)
   (pages :initarg :pages :initform nil :accessor pages)
   (staves :initarg :staves :initform nil :accessor staves)
   (page-height :initarg :page-height :accessor page-height :initform 11.0)
   (page-width :initarg :page-width :accessor page-width :initform 8.5)
   (left-margin :initarg :left-margin :accessor left-margin :initform 0.5)
   (right-margin :initarg :right-margin :accessor right-margin :initform 0.5)
   (header-margin :initarg :header-margin :accessor header-margin :initform 1.0)
   (footer-margin :initarg :footer-margin :accessor footer-margin :initform 1.0)
   (free-expansion-factor :initarg :free-expansion-factor :accessor free-expansion-factor :initform 1.25)
   (line-separation :initarg :line-separation :accessor line-separation :initform 2.0)
   (staff-separation :initarg :staff-separation :accessor staff-separation :initform 1.5)
   (system-separation :initarg :system-separation :accessor system-separation :initform 1.5)
   (size :initarg :size :accessor size :initform 40)
   (beats-per-quarter-note :initarg :beats-per-quarter-note :accessor beats-per-quarter-note :initform 1)
   (clef-breathing-space :initarg :clef-breathing-space :accessor clef-breathing-space :initform .04)
   (leger-line-length :initarg :leger-line-length :accessor leger-line-length :initform .1) ;on the wide side
   (leger-line-width :initarg :leger-line-width :accessor leger-line-width :initform .04)
   (stem-width :initarg :stem-width :accessor stem-width :initform .025)
   (half-stem-width :initarg :half-stem-width :accessor half-stem-width :initform .0125)
   (ideal-stem-length :initarg :ideal-stem-length :accessor ideal-stem-length :initform .5)
   (tie-thickness :initarg :tie-thickness :accessor tie-thickness :initform .04)
   (tie-curvature :initarg :tie-curvature :accessor tie-curvature :initform .125 :accessor tie-curvature)
   (slur-thickness :initarg :slur-thickness :accessor slur-thickness :initform .03)
   (slur-curvature :initarg :slur-curvature :accessor slur-curvature :initform .25 :accessor slur-curvature)
   (maximum-slur-slope :initarg :maximum-slur-slope :accessor maximum-slur-slope :initform .75 :accessor maximum-slur-slope)
   (tie-breathing-space :initarg :tie-breathing-space :accessor tie-breathing-space :initform .04)
   (maximum-beam-tilt :initarg :maximum-beam-tilt :accessor maximum-beam-tilt :initform .4)
   (maximum-subdivision-bracket-tilt :initarg :maximum-subdivision-bracket-tilt :accessor maximum-subdivision-bracket-tilt :initform 0)
   (partial-beam-length :initarg :partial-beam-length :accessor partial-beam-length :initform .4)
   (partial-stem-length :initarg :partial-stem-length :accessor partial-stem-length :initform .10)
   (initial-octave :initarg :initial-octave :accessor initial-octave :initform 4)
   (metronome :initarg :metronome :accessor metronome :initform 60)
   (beam-spacing :initarg :beam-spacing :accessor beam-spacing :initform (* 2.25 .126))
   (beam-width :initarg :beam-width :accessor beam-width :initform (* 1.25 .126))
   (beam-slope-trigger :initarg :beam-slope-trigger :accessor beam-slope-trigger :initform .5)
   (staff-line-width :initarg :staff-line-width :accessor staff-line-width :initform .01)
   (note-head-width :initarg :note-head-width :accessor note-head-width :initform .3)
   (note-head-size :initarg :note-head-size :accessor note-head-size :initform 1.0)
   (staff-line-separation :initarg :staff-line-separation :accessor staff-line-separation :initform .126)
   (half-staff-line-separation :initarg :half-staff-line-separation :accessor half-staff-line-separation :initform (* .5 .126))
   (staff-dy :initarg :staff-dy :accessor staff-dy :initform 1.018)
   (half-staff-dy :initarg :half-staff-dy :accessor half-staff-dy :initform (* .5 1.018))
   (octave-width :initarg :octave-width :accessor octave-width :initform (* 7 .126))
   (sixth-width :initarg :sixth-width :accessor sixth-width :initform (* 5 .126))
   (double-bar-thickness :initarg :double-bar-thickness :initform .1 :accessor double-bar-thickness)
   (double-bar-space :initarg :double-bar-space :initform .1 :accessor double-bar-space)
   (brace-space :initarg :brace-space :initform .125 :accessor brace-space)
   (accidental-breathing-space :initarg :accidental-breathing-space :initform .1 :accessor accidental-breathing-space)
   (dot-distance :initarg :dot-distance :initform .1 :accessor dot-distance)
   (flagged-stem-length :initarg :flagged-stem-length :initform .5 :accessor flagged-stem-length)
   (flag-to-dot-distance :initarg :flag-to-dot-distance :initform .1 :accessor flag-to-dot-distance)
   (curvy-flags :initarg :curvy-flags :initform t :accessor curvy-flags)
   (old-style-beams :initarg :old-style-beams :initform nil :accessor old-style-beams)
   (automatic-page-numbers :initarg :automatic-page-numbers :initform nil :accessor automatic-page-numbers)
   (flag-dy :initarg :flag-dy :initform .2 :accessor flag-dy)
   (flag-up-y-offset :initarg :flag-up-y-offset :initform #+Sonata -.475 #+Petrucci-Medium .125 :accessor flag-up-y-offset)
   (flag-down-y-offset :initarg :flag-down-y-offset :initform #+Sonata .45 #+Petrucci-Medium -.25 :accessor flag-down-y-offset)
   (note-breathing-space :initarg :note-breathing-space :initform .025 :accessor note-breathing-space)
   (leger-line-breathing-space :initarg :leger-line-breathing-space :initform .05 :accessor leger-line-breathing-space)
   (use-italian-octave-signs :initarg :use-italian-octave-signs :initform nil :accessor use-italian-octave-signs)
   (add-bassa-to-octave-signs :initarg :add-bassa-to-octave-signs :initform nil :accessor add-bassa-to-octave-signs)
   (octave-sign-vertical-separation :initarg :octave-sign-vertical-separation :initform 3 :accessor octave-sign-vertical-separation)
   (octave-sign-font :initarg :octave-sign-font :initform "Times-Roman" :accessor octave-sign-font)
   (octave-sign-size :initarg :octave-sign-size :initform .5 :accessor octave-sign-size)
   (octave-sign-pattern :initarg :octave-sign-pattern :initform '(10 20) :accessor octave-sign-pattern)
   (partial-beam-limit :initarg :partial-beam-limit :initform .4 :accessor partial-beam-limit)
   (dynamics-font-scaler :initarg :dynamics-font-scaler :initform .8 :accessor dynamics-font-scaler)
   (marcato-thickness :initarg :marcato-thickness :initform .025 :accessor marcato-thickness)
   (bartok-pizzicato-thickness :initarg :bartok-pizzicato-thickness :initform .025 :accessor bartok-pizzicato-thickness)
   (beat-subdivision-font :initarg :beat-subdivision-font :initform  "Times-Roman" :accessor beat-subdivision-font)
   (beat-subdivision-font-scaler :initarg :beat-subdivision-font-scaler :initform 0.5 :accessor beat-subdivision-font-scaler)
   (rehearsal-font :initarg :rehearsal-font :initform  "Times-Bold" :accessor rehearsal-font)
   (rehearsal-font-scaler :initarg :rehearsal-font-scaler :initform 1.0 :accessor rehearsal-font-scaler)
   (rehearsal-frame-width :initarg :rehearsal-frame-width :initform .075 :accessor rehearsal-frame-width)
   (measure-number-font :initarg :measure-number-font :initform  "Times-Italic" :accessor measure-number-font)
   (measure-number-font-scaler :initarg :measure-number-font-scaler :initform 0.5 :accessor measure-number-font-scaler)
   (repeat-font :initarg :repeat-font :initform  "Times-Roman" :accessor repeat-font)
   (repeat-font-scaler :initarg :repeat-font-scaler :initform .5 :accessor repeat-font-scaler)
   (text-font :initarg :text-font :initform  "Times-Roman" :accessor text-font)
   (tremolo-font :initarg :tremolo-font :initform  "Times-Roman" :accessor tremolo-font)
   (tremolo-font-scaler :initarg :tremolo-font-scaler :initform .4 :accessor tremolo-font-scaler)
   (tremolo-thickness :initarg :tremolo-thickness :initform .08 :accessor tremolo-thickness)
   (tremolo-maximum-length :initarg :tremolo-maximum-length :initform .85 :accessor tremolo-maximum-length)
   (tremolo-slant :initarg :tremolo-slant :initform .1 :accessor tremolo-slant)
   (tremolo-width :initarg :tremolo-width :initform .3 :accessor tremolo-width)
   (pause-font :initarg :pause-font :initform "Times-Bold" :accessor pause-font)
   (staff-name-font :initarg :staff-name-font :initform "Times-Roman" :accessor staff-name-font)
   (staff-name-font-scaler :initarg :staff-name-font-scaler :initform .4 :accessor staff-name-font-scaler)
   (staff-name-font-minimum-size :initarg :staff-name-font-minimum-size :initform 8 :accessor staff-name-font-minimum-size)
   (use-abbreviated-staff-names :initarg :use-abbreviated-staff-names :initform t :accessor use-abbreviated-staff-names)
   (grace-note-size :initarg :grace-note-size :initform .5 :accessor grace-note-size)
   (grace-note-slash-thickness :initarg :grace-note-slash-thickness :initform .025 :accessor grace-note-slash-thickness)
   (text-connecting-pattern :initarg :text-connecting-pattern :initform '(10 0) :accessor text-connecting-pattern)
   (crescendo-thickness :initarg :crescendo-thickness :initform .25 :accessor crescendo-thickness)
   (ending-font :initarg :ending-font :initform "Times-Roman" :accessor ending-font)
   (ending-font-scaler :initarg :ending-font-scaler :initform .4 :accessor ending-font-scaler)
   (always-show-staff-names :initarg :always-show-staff-names :initform t :accessor always-show-staff-names)
   (all-output-in-one-file :initarg :all-output-in-one-file :initform nil :accessor all-output-in-one-file)
   (automatic-line-breaks :initarg :automatic-line-breaks :initform t :accessor automatic-line-breaks)
   (automatic-octave-signs :initarg :automatic-octave-signs :initform t :accessor automatic-octave-signs)
   (automatic-beams :initarg :automatic-beams :initform t :accessor automatic-beams)
   (liberal-automatic-beams :initarg :liberal-automatic-beams :initform t :accessor liberal-automatic-beams)
   (automatic-ties :initarg :automatic-ties :initform t :accessor automatic-ties)
   (automatic-bars :initarg :automatic-bars :initform t :accessor automatic-bars)
   (automatic-rests :initarg :automatic-rests :initform t :accessor automatic-rests)
   ;;(automatic-clefs :initarg :automatic-clefs :initform t :accessor automatic-clefs)
   (automatic-naturals :initarg :automatic-naturals :initform nil :accessor automatic-naturals)
   (implicit-accidental-style :initarg :implicit-accidental-style :initform :new-style :accessor implicit-accidental-style)
   (implicit-accidental-duration :initarg :implicit-accidental-duration :initform nil :accessor implicit-accidental-duration)
   (automatic-beat-subdivision-numbers :initarg :automatic-beat-subdivision-numbers :initform t :accessor automatic-beat-subdivision-numbers)
   (show-rulers :initarg :show-rulers :initform nil :accessor show-rulers)
   (initial-onset :initarg :initial-onset :initform 0 :accessor initial-onset)
   (full-last-line :initarg :full-last-line :initform nil :accessor full-last-line)
   (fingering-font :initarg :fingering-font :initform  "Times-Roman" :accessor fingering-font)
   (fingering-font-scaler :initarg :fingering-font-scaler :initform .33 :accessor fingering-font-scaler)))

(self-action title setf-title)
(self-action page-height setf-page-height)
(self-action page-width setf-page-width)
(self-action left-margin setf-left-margin)
(self-action right-margin setf-right-margin)
(self-action header-margin setf-header-margin)
(self-action footer-margin setf-footer-margin)
(self-action free-expansion-factor setf-free-expansion-factor)
(self-action line-separation setf-line-separation)
(self-action staff-separation setf-staff-separation)
(self-action system-separation setf-system-separation)

					
;;;(self-action size setf-size)
;;;  here we need to know as fast as possible what the score size is
(defgeneric size (val) )
(defmethod size ((val number)) 
  (if (not *cmn-score-size*) (setf *cmn-score-size* val))
  (make-self-acting :action #'setf-size :arguments val))
(defmethod setf-size ((obj score) val) (setf (size obj) val))


;;;(self-action output-file setf-output-file)
;;;  in very large scores, cmn can take a long time before trying to open the output file,
;;;  which is annoying if it dies in the attempt -- output-file therefore makes an
;;;  immediate attempt to open the file so that cmn dies right away if it's going to anyway.
(defgeneric output-file (name) )
(defmethod output-file (name)
  (let ((full-name name)
	(fil nil))
    (if (not (pathname-type full-name)) 
	(setf full-name (concatenate 'string full-name ".eps")))
    (setf fil (open full-name :direction :output :if-exists :supersede))
    (close fil)
    (make-self-acting :action #'setf-output-file :arguments name)))
(defmethod setf-output-file ((obj score) name) (setf (output-file obj) name))

(self-action beats-per-quarter-note setf-beats-per-quarter-note)
(self-action clef-breathing-space setf-clef-breathing-space)
(self-action leger-line-length setf-leger-line-length)
(self-action leger-line-width setf-leger-line-width)
(self-action stem-width setf-stem-width)
(self-action half-stem-width setf-half-stem-width)
(self-action ideal-stem-length setf-ideal-stem-length)
(self-action tie-thickness setf-tie-thickness)
(self-action tie-curvature setf-tie-curvature)
(self-action slur-thickness setf-slur-thickness)
(self-action slur-curvature setf-slur-curvature)
(self-action maximum-slur-slope setf-maximum-slur-slope)
(self-action tie-breathing-space setf-tie-breathing-space)
(self-action maximum-beam-tilt setf-maximum-beam-tilt)
(self-action maximum-subdivision-bracket-tilt setf-maximum-subdivision-bracket-tilt)
(self-action partial-beam-length setf-partial-beam-length)
(self-action partial-stem-length setf-partial-stem-length)
(self-action initial-octave setf-initial-octave)
(self-action metronome setf-metronome)
(self-action beam-spacing setf-beam-spacing)
(self-action beam-width setf-beam-width)
(self-action beam-slope-trigger setf-beam-slope-trigger)
(self-action staff-line-width setf-staff-line-width)
(self-action note-head-width setf-note-head-width)
(self-action note-head-size setf-note-head-size)
(self-action staff-line-separation setf-staff-line-separation)
(self-action half-staff-line-separation setf-half-staff-line-separation)
(self-action staff-dy setf-staff-dy)
(self-action half-staff-dy setf-half-staff-dy)
(self-action octave-width setf-octave-width)
(self-action sixth-width setf-sixth-width)
(self-action double-bar-thickness setf-double-bar-thickness)
(self-action double-bar-space setf-double-bar-space)
(self-action brace-space setf-brace-space)
(self-action accidental-breathing-space setf-accidental-breathing-space)
(self-action dot-distance setf-dot-distance)
(self-action flagged-stem-length setf-flagged-stem-length)
(self-action flag-to-dot-distance setf-flag-to-dot-distance)
(self-action curvy-flags setf-curvy-flags)
(self-action old-style-beams setf-old-style-beams)
(self-action automatic-page-numbers setf-automatic-page-numbers)
(self-action flag-dy setf-flag-dy)
(self-action flag-up-y-offset setf-flag-up-y-offset)
(self-action flag-down-y-offset setf-flag-down-y-offset)
(self-action note-breathing-space setf-note-breathing-space)
(self-action leger-line-breathing-space setf-leger-line-breathing-space)
(self-action partial-beam-limit setf-partial-beam-limit)
(self-action use-italian-octave-signs setf-use-italian-octave-signs)
(self-action add-bassa-to-octave-signs setf-add-bassa-to-octave-signs)
(self-action octave-sign-vertical-separation setf-octave-sign-vertical-separation)
(self-action octave-sign-font setf-octave-sign-font)
(self-action octave-sign-size setf-octave-sign-size)
(self-action octave-sign-pattern setf-octave-sign-pattern)
(self-action dynamics-font-scaler setf-dynamics-font-scaler)
(self-action marcato-thickness setf-marcato-thickness)
(self-action bartok-pizzicato-thickness setf-bartok-pizzicato-thickness)
(self-action repeat-font setf-repeat-font)
(self-action repeat-font-scaler setf-repeat-font-scaler)
(self-action beat-subdivision-font setf-beat-subdivision-font)
(self-action beat-subdivision-font-scaler setf-beat-subdivision-font-scaler)
(self-action rehearsal-font setf-rehearsal-font)
(self-action rehearsal-font-scaler setf-rehearsal-font-scaler)
(self-action rehearsal-frame-width setf-rehearsal-frame-width)
(self-action measure-number-font setf-measure-number-font)
(self-action measure-number-font-scaler setf-measure-number-font-scaler)
(self-action text-font setf-text-font)
(self-action tremolo-font setf-tremolo-font)
(self-action tremolo-font-scaler setf-tremolo-font-scaler)
(self-action tremolo-thickness setf-tremolo-thickness)
(self-action tremolo-maximum-length setf-tremolo-maximum-length)
(self-action tremolo-slant setf-tremolo-slant)
(self-action tremolo-width setf-tremolo-width)
(self-action pause-font setf-pause-font)
(self-action staff-name-font setf-staff-name-font)
(self-action staff-name-font-scaler setf-staff-name-font-scaler)
(self-action staff-name-font-minimum-size setf-staff-name-font-minimum-size)
(self-action use-abbreviated-staff-names setf-use-abbreviated-staff-names)
(self-action grace-note-size setf-grace-note-size)
(self-action grace-note-slash-thickness setf-grace-note-slash-thickness)
(self-action text-connecting-pattern setf-text-connecting-pattern)
(self-action crescendo-thickness setf-crescendo-thickness)
(self-action ending-font setf-ending-font)
(self-action ending-font-scaler setf-ending-font-scaler)
(self-action always-show-staff-names setf-always-show-staff-names)
(self-action all-output-in-one-file setf-all-output-in-one-file)
(self-action automatic-line-breaks setf-automatic-line-breaks)
(self-action automatic-octave-signs setf-automatic-octave-signs)
(self-action automatic-beams setf-automatic-beams)
(self-action liberal-automatic-beams setf-liberal-automatic-beams)
(self-action automatic-ties setf-automatic-ties)
(self-action automatic-bars setf-automatic-bars)
(self-action automatic-rests setf-automatic-rests)
;;(self-action automatic-clefs setf-automatic-clefs)
(self-action automatic-naturals setf-automatic-naturals)
(self-action implicit-accidental-style setf-implicit-accidental-style)
(self-action implicit-accidental-duration setf-implicit-accidental-duration)
(self-action automatic-beat-subdivision-numbers setf-automatic-beat-subdivision-numbers)
(self-action show-rulers setf-show-rulers)
(self-action initial-onset setf-initial-onset)
(self-action full-last-line setf-full-last-line)
(self-action fingering-font setf-fingering-font)
(self-action fingering-font-scaler setf-fingering-font-scaler)


(defmethod copy ((score score) &optional object)
  (if (not object)
      (make-instance 'score 
	  :systems (loop for system in (systems score) collect (copy system))
	  :title (title score) 
	  :time-line (copy-tree (time-line score))
	  :time-lines (copy-tree (time-lines score))
	  :line-data (copy-tree (line-data score))
	  :lines (lines score)
	  :pages (pages score)
	  :staves (staves score)
	  :size (size score) 
	  :page-height (page-height score)
	  :page-width (page-width score)
	  :left-margin (left-margin score)
	  :right-margin (right-margin score)
	  :header-margin (header-margin score)
	  :footer-margin (footer-margin score)
	  :free-expansion-factor (free-expansion-factor score)
	  :line-separation (line-separation score)
	  :staff-separation (staff-separation score)
	  :system-separation (system-separation score)
	  :size (size score)
	  :beats-per-quarter-note (beats-per-quarter-note score)
	  :clef-breathing-space (clef-breathing-space score)
	  :leger-line-length (leger-line-length score)
	  :leger-line-width (leger-line-width score)
	  :stem-width (stem-width score)
	  :half-stem-width (half-stem-width score)
	  :ideal-stem-length (ideal-stem-length score)
	  :tie-thickness (tie-thickness score)
	  :tie-curvature (tie-curvature score)
	  :slur-thickness (slur-thickness score)
	  :slur-curvature (slur-curvature score)
	  :maximum-slur-slope (maximum-slur-slope score)
	  :tie-breathing-space (tie-breathing-space score)
	  :maximum-beam-tilt (maximum-beam-tilt score)
	  :maximum-subdivision-bracket-tilt (maximum-subdivision-bracket-tilt score)
	  :partial-beam-length (partial-beam-length score)
	  :partial-stem-length (partial-stem-length score)
	  :initial-octave (initial-octave score)
	  :metronome (metronome score)
	  :beam-spacing (beam-spacing score)
	  :beam-width (beam-width score)
	  :beam-slope-trigger (beam-slope-trigger)
	  :staff-line-width (staff-line-width score)
	  :note-head-width (note-head-width score)
	  :note-head-size (note-head-size score)
	  :staff-line-separation (staff-line-separation score)
	  :half-staff-line-separation (half-staff-line-separation score)
	  :staff-dy (staff-dy score)
	  :half-staff-dy (half-staff-dy score)
	  :octave-width (octave-width score)
	  :sixth-width (sixth-width score)
	  :double-bar-thickness (double-bar-thickness score)
	  :double-bar-space (double-bar-space score)
	  :brace-space (brace-space score)
	  :accidental-breathing-space (accidental-breathing-space score)
	  :dot-distance (dot-distance score)
	  :flagged-stem-length (flagged-stem-length score)
	  :flag-to-dot-distance (flag-to-dot-distance score)
	  :old-style-beams (old-style-beams score)
	  :automatic-page-numbers (automatic-page-numbers score)
	  :curvy-flags (curvy-flags score)
	  :flag-dy (flag-dy score)
	  :flag-up-y-offset (flag-up-y-offset score)
	  :flag-down-y-offset (flag-down-y-offset score)
	  :note-breathing-space (note-breathing-space score)
	  :leger-line-breathing-space (leger-line-breathing-space score)
	  :use-italian-octave-signs (use-italian-octave-signs score)
	  :add-bassa-to-octave-signs (add-bassa-to-octave-signs score)
	  :octave-sign-vertical-separation (octave-sign-vertical-separation score)
	  :octave-sign-font (octave-sign-font score)
	  :octave-sign-size (octave-sign-size score)
	  :octave-sign-pattern (octave-sign-pattern score)
	  :partial-beam-limit (partial-beam-limit score)
	  :dynamics-font-scaler (dynamics-font-scaler score)
	  :marcato-thickness (marcato-thickness score)
	  :bartok-pizzicato-thickness (bartok-pizzicato-thickness score)
	  :repeat-font (repeat-font score)
	  :repeat-font-scaler (repeat-font-scaler score)
	  :beat-subdivision-font (beat-subdivision-font score)
	  :beat-subdivision-font-scaler (beat-subdivision-font-scaler score)
	  :rehearsal-font (rehearsal-font score)
	  :rehearsal-font-scaler (rehearsal-font-scaler score)
	  :rehearsal-frame-width (rehearsal-frame-width score)
	  :measure-number-font (measure-number-font score)
	  :measure-number-font-scaler (measure-number-font-scaler score)
	  :text-font (text-font score)
	  :tremolo-font (tremolo-font score)
	  :tremolo-font-scaler (tremolo-font-scaler score)
	  :tremolo-thickness (tremolo-thickness score)
	  :tremolo-maximum-length (tremolo-maximum-length score)
	  :tremolo-slant (tremolo-slant score)
	  :tremolo-width (tremolo-width score)
	  :pause-font (pause-font score)
	  :staff-name-font (staff-name-font score)
	  :staff-name-font-scaler (staff-name-font-scaler score)
	  :staff-name-font-minimum-size (staff-name-font-minimum-size score)
	  :use-abbreviated-staff-names (use-abbreviated-staff-names score)
	  :grace-note-size (grace-note-size score)
	  :grace-note-slash-thickness (grace-note-slash-thickness score)
	  :text-connecting-pattern (text-connecting-pattern score)
	  :crescendo-thickness (crescendo-thickness score)
	  :ending-font (ending-font score)
	  :ending-font-scaler (ending-font-scaler score)
	  :always-show-staff-names (always-show-staff-names score)
	  :all-output-in-one-file (all-output-in-one-file score)
	  :automatic-line-breaks (automatic-line-breaks score)
	  :automatic-octave-signs (automatic-octave-signs score)
	  :automatic-beams (automatic-beams score)
	  :liberal-automatic-beams (liberal-automatic-beams score)
	  :automatic-ties (automatic-ties score)
	  :automatic-bars (automatic-bars score)
	  :automatic-rests (automatic-rests score)
	  ;;:automatic-clefs (automatic-clefs score)
	  :automatic-naturals (automatic-naturals score)
	  :implicit-accidental-style (implicit-accidental-style score)
	  :implicit-accidental-duration (implicit-accidental-duration score)
	  :automatic-beat-subdivision-numbers (automatic-beat-subdivision-numbers score)
	  :show-rulers (show-rulers score)
	  :initial-onset (initial-onset score)
	  :full-last-line (full-last-line score)
	  :fingering-font (fingering-font score)
	  :fingering-font-scaler (fingering-font-scaler score))
    (error "can't copy embedded scores yet")))


(defun print-if-changed (field field-name init-value &optional store-form)
  (let ((init-char (if store-form "(" ":"))
	(end-char (if store-form ")" "")))
    (or (and (not init-value)
	     field
	     (format nil "~%  ~A~A ~A~A" init-char field-name field end-char))
	(and init-value
	     (or (and (eq init-value t)
		      (not field)
		      (format nil "~%  ~A~A nil~A" init-char field-name end-char))
		 (and (numberp init-value)
		      (/= init-value field)
		      (format nil "~%  ~A~A ~A~A" init-char field-name field end-char))
		 (and (listp init-value)
		      (not (equal init-value field))
		      (format nil "~%  ~A~A '~A~A" init-char field-name field end-char))
		 (and (stringp init-value)
		      (not (string-equal init-value field))
		      (format nil "~%  ~A~A ~S~A" init-char field-name field end-char))
		 ""))
	"")))

(defun style-list (score &optional (all t) (sf nil))
  (list 
   (if all (print-if-changed (time-lines score) "time-lines" nil sf) "")
   (if all (print-if-changed (line-data score) "line-data" nil sf) "")
   (if all (print-if-changed (lines score) "lines" nil sf) "")
   (if all (print-if-changed (pages score) "pages" nil sf) "")
   (if all (print-if-changed (staves score) "staves" nil sf) "")
   (print-if-changed (page-height score) "page-height" 11.0 sf)
   (print-if-changed (page-width score) "page-width" 8.5 sf)
   (print-if-changed (left-margin score) "left-margin" 0.5 sf)
   (print-if-changed (right-margin score) "right-margin" 0.5 sf)
   (print-if-changed (header-margin score) "header-margin" 1.0 sf)
   (print-if-changed (footer-margin score) "footer-margin" 1.0 sf)
   (print-if-changed (free-expansion-factor score) "free-expansion-factor" 1.25 sf)
   (print-if-changed (line-separation score) "line-separation" 2.0 sf)
   (print-if-changed (staff-separation score) "staff-separation" 1.5 sf)
   (print-if-changed (system-separation score) "system-separation" 1.5 sf)
   (print-if-changed (size score) "size" 40 sf)
   (print-if-changed (beats-per-quarter-note score) "beats-per-quarter-note" 1 sf)
   (print-if-changed (clef-breathing-space score) "clef-breathing-space" .04 sf)
   (print-if-changed (leger-line-length score) "leger-line-length" .1 sf) ;on the wide side
   (print-if-changed (leger-line-width score) "leger-line-width" .04 sf)
   (print-if-changed (stem-width score) "stem-width" .025 sf)
   (print-if-changed (half-stem-width score) "half-stem-width" .0125 sf)
   (print-if-changed (ideal-stem-length score) "ideal-stem-length" .5 sf)
   (print-if-changed (tie-thickness score) "tie-thickness" .04 sf)
   (print-if-changed (tie-curvature score) "tie-curvature" .125 sf)
   (print-if-changed (slur-thickness score) "slur-thickness" .03 sf)
   (print-if-changed (slur-curvature score) "slur-curvature" .25 sf)
   (print-if-changed (maximum-slur-slope score) "maximum-slur-slope" .75 sf)
   (print-if-changed (tie-breathing-space score) "tie-breathing-space" .04 sf)
   (print-if-changed (maximum-beam-tilt score) "maximum-beam-tilt" .4 sf)
   (print-if-changed (maximum-subdivision-bracket-tilt score) "maximum-subdivision-bracket-tilt" 0 sf)
   (print-if-changed (partial-beam-length score) "partial-beam-length" .4 sf)
   (print-if-changed (partial-stem-length score) "partial-stem-length" .10 sf)
   (print-if-changed (initial-octave score) "initial-octave" 4 sf)
   (print-if-changed (metronome score) "metronome" 60 sf)
   (print-if-changed (beam-spacing score) "beam-spacing" (* 2.25 .126) sf)
   (print-if-changed (beam-width score) "beam-width" (* 1.25 .126) sf)
   (print-if-changed (beam-slope-trigger score) "beam-slope-trigger" .5 sf)
   (print-if-changed (staff-line-width score) "staff-line-width" .01 sf)
   (print-if-changed (note-head-width score) "note-head-width" .3 sf)
   (print-if-changed (note-head-size score) "note-head-size" 1.0 sf)
   (print-if-changed (staff-line-separation score) "staff-line-separation" .126 sf)
   (print-if-changed (half-staff-line-separation score) "half-staff-line-separation" (* .5 .126) sf)
   (print-if-changed (staff-dy score) "staff-dy" 1.018 sf)
   (print-if-changed (half-staff-dy score) "half-staff-dy" (* .5 1.018) sf)
   (print-if-changed (octave-width score) "octave-width" (* 7 .126) sf)
   (print-if-changed (sixth-width score) "sixth-width" (* 5 .126) sf)
   (print-if-changed (double-bar-thickness score) "double-bar-thickness" .1 sf)
   (print-if-changed (double-bar-space score) "double-bar-space" .1 sf)
   (print-if-changed (brace-space score) "brace-space" .125 sf)
   (print-if-changed (accidental-breathing-space score) "accidental-breathing-space" .1 sf)
   (print-if-changed (dot-distance score) "dot-distance" .1 sf)
   (print-if-changed (flagged-stem-length score) "flagged-stem-length" .5 sf)
   (print-if-changed (flag-to-dot-distance score) "flag-to-dot-distance" .1 sf)
   (print-if-changed (curvy-flags score) "curvy-flags" t sf)
   (print-if-changed (old-style-beams score) "old-style-beams" nil sf)
   (print-if-changed (automatic-page-numbers score) "automatic-page-numbers" nil sf)
   (print-if-changed (flag-dy score) "flag-dy" .2 sf)
   (print-if-changed (flag-up-y-offset score) "flag-up-y-offset" -.475 sf)
   (print-if-changed (flag-down-y-offset score) "flag-down-y-offset" .45 sf)
   (print-if-changed (note-breathing-space score) "note-breathing-space" .025 sf)
   (print-if-changed (leger-line-breathing-space score) "leger-line-breathing-space" .05 sf)
   (print-if-changed (use-italian-octave-signs score) "use-italian-octave-signs" nil sf)
   (print-if-changed (add-bassa-to-octave-signs score) "add-bassa-to-octave-signs" nil sf)
   (print-if-changed (octave-sign-vertical-separation score) "octave-sign-vertical-separation" 3 sf)
   (print-if-changed (octave-sign-font score) "octave-sign-font" "Times-Roman" sf)
   (print-if-changed (octave-sign-size score) "octave-sign-size" .5 sf)
   (print-if-changed (octave-sign-pattern score) "octave-sign-pattern" '(10 20) sf)
   (print-if-changed (dynamics-font-scaler score) "dynamics-font-scaler" .8 sf)
   (print-if-changed (marcato-thickness score) "marcato-thickness" .025 sf)
   (print-if-changed (bartok-pizzicato-thickness score) "bartok-pizzicato-thickness" .025 sf)
   (print-if-changed (repeat-font score) "repeat-font" "Times-Roman" sf)
   (print-if-changed (repeat-font-scaler score) "repeat-font-scaler" .5 sf)
   (print-if-changed (beat-subdivision-font score) "beat-subdivision-font" "Times-Roman" sf)
   (print-if-changed (beat-subdivision-font-scaler score) "beat-subdivision-font-scaler" 0.5 sf)
   (print-if-changed (rehearsal-font score) "rehearsal-font" "Times-Bold" sf)
   (print-if-changed (rehearsal-font-scaler score) "rehearsal-font-scaler" 1.0 sf)
   (print-if-changed (rehearsal-frame-width score) "rehearsal-frame-width" .075 sf)
   (print-if-changed (measure-number-font score) "measure-number-font" "Times-Italic" sf)
   (print-if-changed (measure-number-font-scaler score) "measure-number-font-scaler" 0.5 sf)
   (print-if-changed (text-font score) "text-font" "Times-Roman" sf)
   (print-if-changed (tremolo-font score) "tremolo-font" "Times-Roman" sf)
   (print-if-changed (tremolo-font-scaler score) "tremolo-font-scaler" .4 sf)
   (print-if-changed (tremolo-thickness score) "tremolo-thickness" .08 sf)
   (print-if-changed (tremolo-maximum-length score) "tremolo-maximum-length" .85 sf)
   (print-if-changed (tremolo-slant score) "tremolo-slant" .1 sf)
   (print-if-changed (tremolo-width score) "tremolo-width" .3 sf)
   (print-if-changed (fingering-font score) "fingering-font" "Times-Roman" sf)
   (print-if-changed (pause-font score) "pause-font" "Times-Bold" sf)
   (print-if-changed (staff-name-font score) "staff-name-font" "Times-Roman" sf)
   (print-if-changed (staff-name-font-scaler score) "staff-name-font-scaler" .4 sf)
   (print-if-changed (staff-name-font-minimum-size score) "staff-name-font-minimum-size" 8 sf)
   (print-if-changed (use-abbreviated-staff-names score) "use-abbreviated-staff-names" t sf)
   (print-if-changed (grace-note-size score) "grace-note-size" .5 sf)
   (print-if-changed (grace-note-slash-thickness score) "grace-note-slash-thickness" .025 sf)
   (print-if-changed (text-connecting-pattern score) "text-connecting-pattern" '(10 0) sf)
   (print-if-changed (crescendo-thickness score) "crescendo-thickness" .25 sf)
   (print-if-changed (ending-font score) "ending-font" "Times-Roman" sf)
   (print-if-changed (ending-font-scaler score) "ending-font-scaler" .4 sf)
   (print-if-changed (always-show-staff-names score) "always-show-staff-names" t sf)
   (print-if-changed (all-output-in-one-file score) "all-output-in-one-file" nil sf)
   (print-if-changed (automatic-line-breaks score) "automatic-line-breaks" t sf)
   (print-if-changed (automatic-octave-signs score) "automatic-octave-signs" t sf)
   (print-if-changed (automatic-beams score) "automatic-beams" t sf)
   (print-if-changed (liberal-automatic-beams score) "liberal-automatic-beams" t sf)
   (print-if-changed (automatic-ties score) "automatic-ties" t sf)
   (print-if-changed (automatic-bars score) "automatic-bars" t sf)
   (print-if-changed (automatic-rests score) "automatic-rests" t sf)
   ;;(print-if-changed (automatic-clefs score) "automatic-clefs" t sf)
   (print-if-changed (automatic-naturals score) "automatic-naturals" nil sf)
   (print-if-changed (implicit-accidental-style score) "implicit-accidental-style" :new-style sf)
   (print-if-changed (implicit-accidental-duration score) "implicit-accidental-duration" nil sf)
   (print-if-changed (automatic-beat-subdivision-numbers score) "automatic-beat-subdivision-numbers" t sf)
   (print-if-changed (show-rulers score) "show-rulers" nil sf)
   (print-if-changed (initial-onset score) "initial-onset" 0 sf)
   (print-if-changed (full-last-line score) "full-last-line" nil sf)
   (print-if-changed (fingering-font-scaler score) "fingering-font-scaler" .33 sf)
   (print-if-changed (partial-beam-limit score) "partial-beam-limit" .4 sf)))

(defmethod descry ((score score) &optional stream controller)
  (format stream "(score ~A~A~A~{~A~}~%  ~A)"
	  (if (title score) (format nil " :title ~A" (title score)) "")
	  (if (next-method-p) (call-next-method score stream (or controller score)) "")
	  (if (time-line score)
	      (format nil "~%   :time-line (list ~{~A~})" (time-line score))
	    "")
	  (style-list score)
	  (if (systems score) (format nil " :systems (list ~{~%     ~A~})" 
				      (loop for system in (systems score) collect (descry system stream score)))
	    "")))

(defmethod score-p ((obj t)) nil)
(defmethod score-p ((obj score)) t)

(defmethod identify ((score score))
  (format nil "score~A" (if (title score) (format nil " titled ~S" (title score)) "")))



(defvar *cmn-system* nil)		;for better error reporting
(defvar *cmn-staff* nil)
(defvar *cmn-staff-data* nil)
(defvar *cmn-object* nil)
(defvar *cmn-owning-object* nil)
(defvar *cmn-page* nil)
(defvar cmn-pipe-0 nil)
(defvar cmn-pipe-1 nil)
(defvar cmn-pipe-2 nil)




;;; method display and house score are at very end -- same for systems?

(defvar no-initialization nil)

(defun scorify (objects)
  (if (and (= (length objects) 1) (score-p (first objects)))
      (setf *cmn-score* (first objects))
    (let ((new-score (make-score)))
      (setf *cmn-score* new-score)
      (multiple-value-bind
	  (systems actions marks)
	  (notify new-score objects)
	(setf (systems new-score) systems)
	(setf (marks new-score) marks)
	(loop for act in actions do
	  (funcall (action act) new-score (arguments act))))))
  (if (title *cmn-score*) 
      (add-to-marks *cmn-score* (list (make-title (title *cmn-score*)))))
#+Excl  (setf (c-output-in-progress *cmn-score*) t)
#-Excl  (setf (c-output-in-progress *cmn-score*) nil)
  (if (not no-initialization) 
      (if (output-file *cmn-score*)
	  (initialize *cmn-score* :with-pages (all-output-in-one-file *cmn-score*))
	(initialize *cmn-score* :file "aaa.eps" :with-pages t)))
  *cmn-score*)

(defmethod notify ((score score) &optional objects)
  ;; make sure everything in the objects list is parcelled out to the correct portion of the score
  (let ((systems nil)
	(actions nil)
	(marks nil)
	(object-list objects)
	(current-default-system nil)
	(current-default-system-data nil))
    (loop while object-list do
      (let ((object (pop object-list)))
	(when object
	  (if (score-object-list-p object)
	      (setf object-list (append (disgorge object) object-list))
	    (if (system-p object)
		(progn
		  (if current-default-system
		      (push (apply #'system (nreverse current-default-system-data)) systems))
		  (setf current-default-system-data nil)
		  (if (write-protected object)
		      (setf current-default-system object)
		    (progn
		      (setf current-default-system nil)
		      (push object systems))))
	      (if (not current-default-system)
		  (if (self-acting-p object)
		      (push object actions)
		    (if (or (text-p object) (sundry-p object) (glyph-list-p object))
			(push objects marks)
		      (progn
			(setf current-default-system (make-write-protected-system))
			(push object current-default-system-data))))
		(progn
		  (if (not current-default-system)
		      (setf current-default-system (make-write-protected-system)))
		  (push object current-default-system-data))))))))
    (if current-default-system-data (push (apply #'system (nreverse current-default-system-data)) systems))
    (values (nreverse systems) (nreverse actions) marks)))


;;;
;;; --------------    overall organization of cmn:
;;;

(defvar tie-stack nil)
(defvar slur-stack nil)
(defvar glissando-stack nil)
(defvar tremolo-stack nil)
(defvar crescendo-stack nil)
(defvar text-stack nil)
(defvar octave-stack nil)
(defvar rehearsal-stack nil)
(defvar beam-stack nil)

(defun cmn (&rest objects)
  (when objects				; if no args, cmn is a no-op
    (unwind-protect
	(catch 'cannot-open-file	; open file error aborts entire process
	  (cmn-finalize			; cleanup, return complete score structure (for subsequent editing)
	   (drawify			; drawify draws the result (normally creates a separate .eps file for each page)
	    (filterify			; filterify filters out "uninteresting" portions of the final score (for page extraction)
	     (slurify			; slurify adds slurs
	      (markify			; markify adds most marks like fingerings, staccato, sF, etc
	       (tieify			; tieify adds ties
		(beamify		; beamify adds beams
		 (justify		; justify adds in white space to make each line and page look pretty
		  (pagify		; pagify breaks the linified form into pages
		   (linify		; linify breaks the compacted form into lines
		    (compactify		; compactify aligns boxes across all systems
		     (boxify		; boxify finds bounding boxes and center points
		      (fillify		; fillify does the first level of filling in missing data (bar lines, rests etc)
		       (scorify		; scorify (via notify) regularizes the input stream of objects
			(edify		; edify filters out all but "interesting" data (i.e. for part extraction, transposition, etc)
			 (cmn-initialize objects)))))))))))))))))
      (progn
	(if (and *cmn-score* 
		 (c-output-in-progress *cmn-score*))
	    (progn
	      (setf (c-output-in-progress *cmn-score*) nil)
	      (c-close)))
	(if (not no-initialization)
	    (setf tie-stack nil
		  slur-stack nil
		  glissando-stack nil
		  crescendo-stack nil
		  tremolo-stack nil
		  text-stack nil
		  rehearsal-stack nil
		  beam-stack nil
		  octave-stack nil
		  cmn-pipe-0 nil
		  cmn-pipe-1 nil
		  cmn-pipe-2 nil
		  *cmn-score-size* nil
		  *old-cmn-score-size* nil
		  *cmn-system* nil
		  *cmn-staff* nil
		  *cmn-staff-data* nil
		  *cmn-page* nil
		  *cmn-object* nil
		  *cmn-owning-object* nil))))))

(defun check-for-left-overs ()
  (if tie-stack 
      (format t "~%begin-tie without matching end-tie:~{~%    ~A~}"
	      (loop for tie in (reverse tie-stack) collect (descry tie))))
  (if slur-stack 
      (format t "~%begin-slur without matching end-slur:~{~%    ~A~}" 
	      (loop for slur in (reverse slur-stack) collect (descry slur))))
  (if glissando-stack 
      (format t "~%begin-glissando without matching end-glissando:~{~%    ~A~}" 
	      (loop for gliss in (reverse glissando-stack) collect (descry gliss))))
  (if tremolo-stack 
      (format t "~%begin-tremolo without matching end-tremolo:~{~%    ~A~}" 
	      (loop for trem in (reverse tremolo-stack) collect (descry trem))))
  (if crescendo-stack 
      (format t "~%begin-crescendo or diminuendo without matching end:~{~%    ~A~}"
	      (loop for cresc in (reverse crescendo-stack) collect (descry cresc))))
  (if octave-stack 
      (if (listp octave-stack)
	  (format t "~%begin-octave without matching end-octave:~{~%    ~A~}" 
		  (loop for oct in (reverse octave-stack) collect (descry oct)))
	(format t "~%begin-octave, but no end-octave")))
  (if beam-stack 
      (format t "~%begin-beam without matching end-beam:~{~%    ~A~}" 
	      (loop for beam in (reverse beam-stack) collect (descry beam)))))

(defvar editor-hook nil)
(defvar filter-hook nil)
(defvar page-hook nil)
(defvar line-hook nil)

(defmethod cmn-finalize ((score score)) 
  (if (not no-initialization)		;might want to save/restore "hooks" across with-cmn
      (progn
	(check-for-left-overs)
	(setf editor-hook nil)
	(setf filter-hook nil)
	(finalize score)))
  score)

(defmethod cmn-initialize (objects) 
  objects)

;;; these two are methods so that callers can customize set-up and shut-down behavior.


;;; this is no longer needed -- cmn-initialize is now a method, so edify is just a sort of :after method

;;; edify is the bottom level editor
;;; it is passed the score (a list of lists of objects) and does what it wants

(defun edify (objects)
  (if editor-hook			;if an editor is present, let it munge the un-regularized data list
      (funcall editor-hook objects)
    objects))


;;; filterify should be done by making drawify a method

;;; filterify takes the finished score description (ready for the printer) and
;;; applies one last filtering pass

(defun filterify (score)
  (if filter-hook
      (funcall filter-hook score)
    score))



(defclass score-object-mixin (dxy-mixin bounds-mixin odb-mixin visible-mixin)
  ())

(defmethod score-object-p ((obj t)) nil)
(defmethod score-object-p ((obj score-object-mixin)) t)

(defclass score-object (score-object-mixin dxy bounds odb visible)
  ())

;;; anything that the score needs to keep track of (during alignment for example) is a score-object




;;;
;;; ----------------    system
;;;
;;; a system is a group of staves with auxiliary annotations (measure numbers, overall tempo indications, etc)


(defclass system-mixin (score-object-mixin)
  ((staves :initarg :staves :initform nil :reader staves)
   (bracketing :initarg :bracketing :initform nil :reader bracketing)))

(defclass system (system-mixin score-object)
  ((staves :accessor staves)
   (bracketing :accessor bracketing)))

(defclass write-protected-system (write-protect system-mixin)
  ())

(defmethod system-p ((obj t)) nil)
(defmethod system-p ((obj system-mixin)) t)

(defvar system (make-instance 'write-protected-system))

(defmethod descry ((system system-mixin) &optional stream controller)
  (format stream "~A~A~A~A~A"
	  "(system"
	  (if (bracketing system) 
	      (format nil " ~A:bracketing ~A"
 		      (if controller (format nil "~%       ") "")
		      (bracketing system))
	    "")
	  (if (staves system)
	      (format nil " ~A:staves (list ~{~%~A~})" 
		      (if controller (format nil "~%       ") "")
		      (loop for staff in (staves system) collect (descry staff stream (or controller system))))
	    "")
	  (if (next-method-p) (call-next-method system stream (or controller system)) "")
	  (if (not controller) ")" "")))

(defmethod copy ((system system-mixin) &optional object)
  (let ((new-system (if (not object) (make-system)
		      (if (write-protected object) (copy object)
			object))))
    (setf (bracketing new-system) (bracketing system))
    (setf (staves new-system) (loop for staff in (staves system) collect (copy staff)))
    (if (next-method-p) (call-next-method system new-system))
    new-system))

(defun system (&rest objects)
  (let ((new-system (make-system)))
    (setf *cmn-system* new-system)
    (multiple-value-bind
	(staves actions marks)
	(notify new-system objects)
      (setf (staves new-system) staves)
      (setf (marks new-system) marks)
      (loop for act in actions do
	(funcall (action act) new-system (arguments act))))
    (setf *cmn-system* nil)
    new-system))

(defmethod notify ((system system) &optional objects)
  (let ((default-staff nil)
	(staves nil)
	(actions nil)
	(marks nil)
	(object-list objects)
	(default-staff-data nil))
    (loop while object-list do
      (let ((object (pop object-list)))
	(when object
	  (if (score-object-list-p object)
	      (setf object-list (append (disgorge object) object-list))
	    (if (and (not default-staff-data)
		     (or (bracket-p object) (brace-p object)))
		(setf (bracketing system) object)
	      (if (staff-p object)
		  (progn
		    (if default-staff
			(push (apply #'staff (nreverse default-staff-data)) staves))
		    (setf default-staff-data nil)
		    (if (write-protected object)
			(setf default-staff object)
		      (progn
			(setf default-staff nil)
			(push object staves))))
		(if (not default-staff)
		    (if (self-acting-p object)
			(push object actions)
		      (if (or (text-p object) (sundry-p object) (glyph-list-p object))
			  (push object marks)
			(progn
			  (setf default-staff (make-write-protected-staff))
			  (push object default-staff-data))))
		  (progn
		    (if (not default-staff)
			(setf default-staff (make-write-protected-staff)))
		    (push object default-staff-data)))))))))
    (if default-staff (push (apply #'staff (nreverse default-staff-data)) staves))
    (values (nreverse staves) (nreverse actions) marks)))



;;;
;;; ----------------    staff 
;;;

(defclass staff-mixin (score-object-mixin)
  ((name :initarg :name :initform nil :reader staff-name)
   (data :initarg :data :initform nil :reader staff-data)
   (size :initarg :size :initform nil :reader staff-size)
   (lines :initarg :lines :initform nil :reader staff-lines)
   (start-line :initarg :start-line :initform 0 :reader staff-start-line)
   (local-brace :initarg :local-brace :initform nil :reader staff-local-brace)
   (staff-name-x0 :initarg :staff-name-x0 :initform nil :reader staff-name-x0)
   (true-staff :initarg :true-staff :initform nil :reader true-staff)
   (back-staff :initarg :back-staff :initform nil :reader back-staff)
   (partial-box :initarg :partial-box :initform nil :reader partial-box)))


(defclass write-protected-staff (write-protect staff-mixin)
  ())

(defclass staff (staff-mixin score-object)
  ((name :accessor staff-name)
   (data :accessor staff-data)
   (size :accessor staff-size)
   (lines :accessor staff-lines)
   (start-line :accessor start-line)
   (local-brace :accessor staff-local-brace)
   (staff-name-x0 :accessor staff-name-x0)
   (true-staff :accessor true-staff)
   (back-staff :accessor back-staff)
   (partial-box :accessor partial-box)))

(self-action staff-name setf-name)
(self-action staff-lines setf-lines)
(self-action staff-size setf-staff-size)
(self-action staff-name-x0 setf-staff-name-x0)
(self-action start-line setf-start-line)

(defmethod staff-y0 ((staff staff-mixin)) (box-y0 staff))
(defmethod (setf staff-y0) (val (staff staff-mixin)) (setf (box-y0 staff) val))

(defmethod setf-x0 :after ((staff staff-mixin) val) 
  (if (partial-box staff) (setf (first (partial-box staff)) val) (setf (partial-box staff) (list val nil nil nil))))
(defmethod setf-x1 :after ((staff staff-mixin) val) 
  (if (partial-box staff) (setf (third (partial-box staff)) val) (setf (partial-box staff) (list nil nil val nil))))
(defmethod setf-y0 :after ((staff staff-mixin) val) 
  (if (partial-box staff) (setf (second (partial-box staff)) val) (setf (partial-box staff) (list nil val nil nil))))
(defmethod setf-y1 :after ((staff staff-mixin) val) 
  (if (partial-box staff) (setf (fourth (partial-box staff)) val) (setf (partial-box staff) (list nil nil nil val))))

(defvar staff (make-instance 'write-protected-staff))

(defun identify-staff (score staff)
  (loop for system in (systems score) and sys-num from 1 do
    (loop for stf in (staves system) and staff-num from 1 do
      (if (eq stf staff)
	  (return-from identify-staff 
	    (format nil "~A in system ~D (staff ~D)" 
		    (or (staff-name staff) "un-named staff")
		    sys-num staff-num))))))

(defmethod descry ((staff staff-mixin) &optional stream controller)
  (let ((prespace (if controller  "        " "")))
    (format stream "~A(staff~A~A~A~A~A~A~A~A~A)"
	    prespace
	    (if (staff-name staff) (format nil " :name ~S" (staff-name staff)) "")
	    (if (staff-name-x0 staff) (format nil " :staff-name-x0 ~1,3F" (staff-name-x0 staff)) "")
	    (if (staff-size staff) (format nil " :size ~A" (staff-size staff)) "")
	    (if (staff-lines staff) (format nil " :lines ~A" (staff-lines staff)) "")
	    (if (and (numberp (staff-start-line staff))
		     (> (staff-start-line staff) 0))
		(format nil " :start-line ~D" (staff-start-line staff)) "")
	    (if (staff-local-brace staff) (format nil " :local-brace ~A" (staff-local-brace staff)) "")
	    (if (true-staff staff) (format nil " :true-staff ~A" (true-staff staff)) "")
	    (if (next-method-p) (call-next-method staff stream (or controller staff)) "")
	    (if (staff-data staff) 
		(format nil "~%~A  :data (list~{~%            ~A~})" 
			prespace
			(loop for object in (staff-data staff) 
			 collect (descry object)))
	      ""))))

(defmethod copy ((staff staff-mixin) &optional object)
  (let ((new-staff (if (not object) (make-staff)
		     (if (write-protected object) (copy object)
		       object))))
    (setf (staff-data new-staff) (loop for obj in (staff-data staff) collect (copy obj)))
    (setf (staff-name new-staff) (staff-name staff))
    (setf (staff-name-x0 new-staff) (staff-name-x0 staff))
    (setf (staff-size new-staff) (staff-size staff))
    (setf (staff-lines new-staff) (staff-lines staff))
    (setf (start-line new-staff) (staff-start-line staff))
    (setf (true-staff new-staff) (true-staff staff))
    (setf (back-staff new-staff) (back-staff staff))
    (if (staff-local-brace staff) (setf (staff-local-brace new-staff) (copy (staff-local-brace staff))))
    (if (next-method-p) (call-next-method staff new-staff))
    new-staff))

(defmethod staff-p ((obj t)) nil)
(defmethod staff-p ((obj staff-mixin)) t)

(defun staff (&rest objects)
  (let ((new-staff (make-staff)))
    (multiple-value-bind
	(data actions local-brace)
	(notify new-staff objects)
      (setf *cmn-staff* new-staff)
      (setf (staff-data new-staff) data)
      (setf *cmn-staff-data* nil)
      (cmn-clear-pipe)
      (setf (staff-local-brace new-staff) local-brace)
      (loop for act in actions do
	(funcall (action act) new-staff (arguments act)))
      (if (staff-name new-staff)
	  (add-to-marks new-staff (list (make-staff-name))))
      (setf *cmn-staff* nil)
      new-staff)))

(defvar all-instrument-names 
    '("piccolo" "flute" "alto flute" "bass flute" 
      "oboe" "oboe d'amore" "english horn"
      "e-flat clarinet" "b-flat clarinet" "a clarinet" "bass clarinet" "e-flat contrabass clarinet" "b-flat contrabass clarinet"
      "bassoon" "contrabassoon"
      "soprano saxophone" "alto saxophone" "tenor saxophone" "baritone saxophone" "bass saxophone"
      "french horn" "b-flat piccolo trumpet" "e-flat trumpet" "d trumpet" "c trumpet" "b-flat trumpet"
      "cornet" "fluegelhorn" "bass trumpet"
      "alto trombone" "tenor trombone" "bass trombone" 
      "baritone horn" "euphonium" "tuba"
      "timpani" "vibraphone" "xylophone" "marimba" "glockenspiel" "chimes"
      "celesta" "piano"
      "violin" "viola" "cello" "contrabass"))

(defvar all-abbreviated-instrument-names 
    '("picc." "fl." "alt. fl." "bs. fl." 
      "ob." "ob. d'am." "e.h."
      "e-flat cl." "b-flat cl." "a cl." "bs. cl" "e-flat cbs. cl." "b-flat cbs. cl."
      "bsn." "cbsn."
      "sop. sax." "alt. sax." "ten. sax." "bar. sax." "bs. sax."
      "hn." "b-flat picc. tpt." "e-flat tpt." "d tpt." "c tpt." "b-flat tpt." "cor." "flhn." "bs. tpt."
      "alt. tbn." "ten. tbn." "bs. tbn."
      "bar." "euph." "tba." 
      "timp." "vibe." "xyl." "mar." "glock." "chm."
      "cel." "pno."
      "vn." "vla." "vc." "cb."))

(defmethod abbreviation-of-instrument-name (name)
  (let ((pos (position name all-instrument-names :test #'string-equal)))
    (or (and pos
	     (nth pos all-abbreviated-instrument-names))
	name)))

(defmethod abbreviation-of-instrument-name ((text text))
  (abbreviation-of-instrument-name (letters text)))




;;; audible class given here to make dumb lisp loader happy -- see cmn2.lisp for details

(defclass audible-mixin ()
  ((slurs :initarg :slurs :initform nil :reader slurs)
   (beams :initarg :beams :initform nil :reader beams)
   (beamed :initarg :beamed :initform nil :reader beamed)
   (octaved :initarg :octaved :initform nil :reader octaved)
   (tremolo :initarg :tremolo :initform nil :reader get-tremolo)
   (ties :initarg :ties :initform nil :reader ties)
   (stem-direction :initarg :stem-direction :initform nil :reader stem-direction :reader audible-stem-direction)
   (stem-end :initarg :stem-end :initform nil :reader stem-end)
   (collision :initarg :collision :initform 0 :reader collision :reader audible-collision)
   (outer-beam :initarg :outer-beam :initform nil :reader outer-beam)
   (store-data :initarg :store-data :initform nil :reader store-data)
   (tie-direction :initarg :tie-direction :initform nil :reader tie-direction :reader audible-tie-direction)
   (slur-direction :initarg :slur-direction :initform nil :reader slur-direction :reader audible-slur-direction)))

(defclass audible (audible-mixin)
  ((slurs :accessor slurs)
   (beams :accessor beams)
   (beamed :accessor beamed)
   (octaved :accessor octaved)
   (tremolo :accessor get-tremolo)
   (ties :accessor ties)
   (stem-direction :accessor stem-direction :accessor audible-stem-direction)
   (stem-end :accessor stem-end)
   (collision :accessor collision :accessor audible-collision)
   (outer-beam :accessor outer-beam)
   (store-data :accessor store-data)
   (tie-direction :accessor tie-direction :accessor audible-tie-direction)
   (slur-direction :accessor slur-direction :accessor audible-slur-direction)))

(defmethod audible-p ((obj t)) nil)
(defmethod audible-p ((obj audible-mixin)) t)

(defun stem-is-up (object) (member (audible-stem-direction object) '(:up :up?)))
(defun stem-is-down (object) (member (audible-stem-direction object) '(:down :down?)))
(defun stem-is-locked (object) (member (audible-stem-direction object) '(:up :down)))
(defun stem-is-up? (object) (if (audible-stem-direction object) (stem-is-up object) (< (head-line object) 4)))
(defun stem-is-down? (object) (not (stem-is-up? object)))

(defun tied-to (staff)
  (make-self-acting
   :action #'(lambda (new-staff old-staff)
	       (setf (true-staff new-staff) old-staff)
	       (setf (staff-size new-staff) (staff-size old-staff))
	       nil)
   :arguments staff))

(defun notifiable (obj1 obj2)
  (or (and (audible-p obj1)
	   (or (rhythm-p obj2)
	       (sundry-p obj2)
	       (pause-p obj2)
	       (text-p obj2)
	       (glyph-list-p obj2)
	       (self-acting-p obj2)
	       (tag-p obj2)
	       (dynamics-p obj2)))
      (and (score-object-p obj1)
	   (sundry-p obj2))))

(defmethod notify ((staff staff-mixin) &optional objects)
  ;; just make sure all unattached (write-protected) objects are repackaged -- for the
  ;; initial cases, place clef key meter name and size onto the current staff object (these can change)
  (let ((last-wp-object nil)
	(last-wp-data nil)
	(actions nil)
	(local-brace nil)
	(object-list objects))
    (setf *cmn-staff-data* nil)
    (setf *cmn-staff* nil)
    (cmn-clear-pipe)
    (loop while object-list do
      (let* ((pobj (pop object-list)))
	(when pobj
	  (let* ((cons-obj (write-protected pobj))
		 (ok-obj (if cons-obj (copy pobj) pobj)))
	    (if (score-object-list-p ok-obj)
		(setf object-list (append (disgorge ok-obj) object-list))
	      (if (or (bracket-p ok-obj) (brace-p ok-obj))
		  (setf local-brace ok-obj)
		(let ((object nil))
		  (if last-wp-object
		      (if (not (notifiable last-wp-object ok-obj))
			  (progn
			    (push (funcall #'notify last-wp-object (nreverse last-wp-data)) *cmn-staff-data*)
			    (setf last-wp-data nil)
			    (setf last-wp-object nil)
			    (if cons-obj 
				(setf last-wp-object ok-obj)
			      (setf object ok-obj)))
			(push ok-obj last-wp-data))
		    (if cons-obj
			(setf last-wp-object ok-obj)
		      (setf object ok-obj)))
		  (when object
		    (if (self-acting-p object)
			(push object actions)
		      (if (or (text-p object) (glyph-list-p object) (sundry-p object))
			  (push (make-self-acting :action #'add-to-marks :arguments (list object)) actions)
			(push object *cmn-staff-data*)))))))))))
    (if last-wp-object (push (funcall #'notify last-wp-object (nreverse last-wp-data)) *cmn-staff-data*))
    (if (or (line-p (first *cmn-staff-data*)) (page-p (first *cmn-staff-data*))) (pop *cmn-staff-data*))
    (values (nreverse *cmn-staff-data*) (nreverse actions) local-brace)))

(defmethod display ((staff staff-mixin) container score &rest rest)
  (declare (ignore container))
  (let ((px0 (+ (box-x0 staff) (dxy-dx staff)))
	(py0 (+ (box-y0 staff) (dxy-dy staff)))
	(x1 (box-x1 staff)))
    (when (or px0 py0 x1)
      ;; we are at x y and want the staff to go to x1.
      ;; we don't use %staff because the scale factor causes the dx-left amount to be nearly unpredictable
      ;; that is, the staff "character"'s origin is not the lower staff line's beginning
      (if (marks staff) (apply #'display-marks staff score rest))
      (let* ((x-back (box-x0 %staff))	;mimic font's staff character
	     (y-down (box-y0 %staff))
	     (line-width (staff-line-width score))
	     (lines (or (staff-lines staff) 5))
	     (sls (* 2 (staff-line-separation score)))
	     (x0 (+ px0 x-back))
	     (y0 (+ py0 y-down (* sls (staff-start-line staff)))))
	(comment score "staff lines")
	(setf (line-width score) line-width)
	(loop for i from 0 below lines do
	  (just-line score x0 y0 x1)
	  (incf y0 sls))		;.2545 -- this is to fit the clef signs -- 1018 total distance
	(simple-draw score)
	(setf (line-width score) 0)
	(let* ((ssize (or *cmn-score-size* (scr-size score)))
	       (x0s (* ssize x0))
	       (x1s (* ssize x1))
	       (y0s (* ssize (+ py0 y-down)))
	       (y1s (* ssize (- y0 sls))))
	  (setf (scr-x score) x1s)
	  (setf (scr-y score) y1s)
	  (setf (scr-x0 score) (min (scr-x0 score) x0s x1s))
	  (setf (scr-y0 score) (min (scr-y0 score) y0s y1s))
	  (setf (scr-x1 score) (max (scr-x1 score) x0s x1s))
	  (setf (scr-y1 score) (max (scr-y1 score) y0s y1s)))))))

(defmethod house ((staff staff-mixin) score)
  ;; no obvious x1 yet (depends on rest of data)
  (declare (ignore score))
  (setf (box-y1 staff) 1.0)		;staff is basically this high
  (setf (center staff) 0.0))
;; staff has no other spacing or justification needs




(defclass score-object-list ()
  ((data :initarg :data :initform nil :accessor data)))

(defmethod score-object-list-p ((obj t)) nil)
(defmethod score-object-list-p ((obj score-object-list)) t)

(defmethod disgorge ((objlist score-object-list))
  (prog1
      (data objlist)
    (setf (data objlist) nil)))

(defmethod engorge (objects)
  (make-score-object-list :data objects))

(defmethod descry ((sob score-object-list) &optional stream controller)
  (format stream "~A~A~A~A"
	  (if (not controller) "(score-object-list" 
	    (if (data sob) (format nil "~%~A" prewhitespace)
	      ""))
	  (if (data sob)
	      (format nil " :data (list ~{ ~A~})" (loop for datum in (data sob) collect (descry datum)))
	    "")
	  (if (next-method-p) (call-next-method sob stream (or controller sob)) "")
	  (if (not controller) ")" "")))

(defmethod copy ((sob score-object-list) &optional object)
  (let ((new-sob (if (not object) (make-score-object-list)
		   (if (write-protected object) (copy object)
		     object))))
    (setf (data new-sob) (loop for datum in (data sob) collect (copy datum)))
    (if (next-method-p) (call-next-method sob new-sob))
    new-sob))

(defmethod display ((score-obj score-object-list) container score &rest rest)
  (when (not (invisible-matrix-p score-obj))
    (let ((sobs (data score-obj))
	  (need-gsave (not (identity-matrix-p score-obj))))
      (when (output score)
	(let ((c-out (c-output-in-progress score)))
	  (when need-gsave
	    (if c-out
		(c-print (format nil " gsave~% [ ~{~A ~}] concat~%" (map 'list #'not-rational (matrix score-obj))))
	      (progn
		(format (output score) " gsave~%")
		(format (output score) " [ ~{~A ~}] concat~%" (map 'list #'not-rational (matrix score-obj))))))
	  (loop for sob in sobs do (apply #'display sob container score rest))
	  (if c-out
	      (progn
		(c-print " show")
		(if need-gsave (c-print " grestore")))
	    (progn
	      (format (output score) " show~%")
	      (if need-gsave (format (output score) " grestore~%")))))))))

(defmethod %rx ((server server) (score-obj score-object-list) &key)
  (if (identity-matrix-p score-obj)
      (loop for object in (data score-obj) sum (%rx server object))
    0))



(defclass glyph-list (score-object score-object-list font-mixin) ())

(defmethod glyph-list-p ((obj t)) nil)
(defmethod glyph-list-p ((obj glyph-list)) t)

(defmethod show ((server server) (glyph-obj glyph-list) &key (dx 0) (dy 0) size &allow-other-keys)
  (when (not (invisible-matrix-p glyph-obj))
    (let ((glyphs (loop for glyph in (data glyph-obj) collect (index glyph)))
	  (pstart "")
	  (ssize (or *cmn-score-size* (scr-size server)))
	  (pend (if (and (zerop dx) (zerop dy)) "show" "ashow")))
      (when (output server)
	(let ((font-case (or size (font-name glyph-obj)))
	      (c-out (c-output-in-progress server)))
	  (when font-case
	    (if c-out (c-print " gsave") (format (output server) " gsave~%"))
	    (if c-out (c-print (format nil " /~A findfont ~D scalefont setfont" 
				       (if size (font server) (font-name glyph-obj))
				       (or size (font-size glyph-obj) (* (font-scaler glyph-obj) ssize))))
	      (format (output server) " /~A findfont ~D scalefont setfont~%" 
		      (if size (font server) (font-name glyph-obj))
		      (or size (font-size glyph-obj) (* (font-scaler glyph-obj) ssize)))))
	  (if (pattern glyph-obj)
	      (multiple-value-bind
		  (qpstart qpend)
		  (funcall (pattern glyph-obj) server glyph-obj)
		(if qpstart (setf pstart qpstart))
		(if qpend (setf pend qpend))))
	  (if (identity-matrix-p glyph-obj)
	      (if (or (not (zerop dx)) (not (zerop dy)))
		  (if c-out
		      (c-print (format nil " ~A ~1,3F ~1,3F (~{\\~D~}) ~A~A" pstart (* dx ssize) (* dy ssize) glyphs pend
				       (if font-case " grestore" "")))
		    (princ (format nil " ~A ~1,3F ~1,3F (~{\\~D~}) ~A~A~%" 
				   pstart
				   (* dx ssize)
				   (* dy ssize)
				   glyphs
				   pend
				   (if font-case " grestore" ""))
			   (output server)))
		(if c-out
		    (c-print (format nil " ~A (~{\\~D~}) ~A~A" pstart glyphs pend (if font-case " grestore" "")))
		  (princ (format nil " ~A (~{\\~D~}) ~A~A~%" pstart glyphs pend (if font-case " grestore" "")) 
			 (output server))))
	    (if c-out
		(c-print (format nil " ~A ~A(~{\\~D~}) [ ~{~A ~}] concat ~A grestore"
				 (if (not font-case) "gsave " "") pstart glyphs (map 'list #'not-rational (matrix glyph-obj)) pend))
	      (princ (format nil
			     " ~A ~A(~{\\~D~}) [ ~{~A ~}] concat ~A grestore~%"
			     (if (not font-case) "gsave " "")
			     pstart
			     glyphs
			     (map 'list #'not-rational (matrix glyph-obj))
			     pend)
		     (output server)))))))
    (%dxy server (%rx server glyph-obj :dx dx :size size) 0)
    (%update-bounds server)))

(defmethod %rx ((server server) (score-obj glyph-list) &key (dx 0) size &allow-other-keys)
  (if (identity-matrix-p score-obj)
      (let ((tsize (or size *cmn-score-size* (scr-size server))))
	(* tsize (loop for object in (data score-obj) sum (+ (g-rx object) dx))))
    0))

(defun decimal-to-octal (n &optional (nn 0) (pow 1)) 
  (if (< n 8) 
      (+ (* n pow) nn) 
    (multiple-value-bind 
	(int frac) 
	(floor n 8) 
      (decimal-to-octal int (+ nn (* frac pow)) (* pow 10)))))

(defun glyphs (&rest objects)
  (apply #'ur-glyphs (make-glyph-list) objects))

(defun %glyphs (&rest objects)
  (apply #'ur-glyphs (make-instance 'glyph-list) objects))

(defun ur-glyphs (obj &rest objects)
  (let ((glfs nil))
    (loop for act in objects do
      (when act
	(if (self-acting-p act)
	    (funcall (action act) obj (arguments act))
	  (if (integerp act)
	      (push (make-glyph :index act) glfs)
	    (if (characterp act)
		(push (make-glyph :index (decimal-to-octal (char-code act))) glfs)
	      (push act glfs))))))
    (setf (data obj) (nreverse glfs))
    obj))

(defmethod display ((text glyph-list) container score &rest rest)
  (if (not (member :just-fooling rest))
      (let* ((audp (and container (audible-p container)))
	     (x-off (+ (box-x0 text) (dxy-dx text) (if audp (box-x0 container) 0)))
	     (y-off (+ (box-y0 text) (dxy-dy text) (if audp (box-y0 container) 0))))
	(if (marks text) (apply #'display-marks text score rest))
	(moveto score x-off y-off)
	(show score text))))





;;; some help in error handing

(defun cmn-tick-pipe (obj)
  (setf cmn-pipe-2 cmn-pipe-1)
  (setf cmn-pipe-1 cmn-pipe-0)
  (setf cmn-pipe-0 obj))

(defun cmn-clear-pipe ()
  (setf cmn-pipe-0 nil)
  (setf cmn-pipe-1 nil)
  (setf cmn-pipe-2 nil))

(defun cmn-pipe-context ()
  (if cmn-pipe-0
      (let ((cmn0 (identify cmn-pipe-0)))
	(format nil " recently evaluated note~P:~A~A~A~A~A~A~%~VT"
		(if cmn-pipe-1 2 1)
		(if cmn-pipe-2 (format nil "~%        ") " ")
		(if cmn-pipe-2 (identify cmn-pipe-2) "")
		(if cmn-pipe-1 (format nil "~%        ") " ")
		(if cmn-pipe-1 (identify cmn-pipe-1) "")
		(if cmn-pipe-1 (format nil "~%        ") " ")
		cmn0
		(length cmn0)))
    ""))

(defun system-context (sys)
  (if sys
      (if (and (systems *cmn-score*) 
	       (> (length (systems *cmn-score*)) 1))
	  (let ((pos (position sys (systems *cmn-score*))))
	    (if pos (format nil "in system ~D " pos)
	      "in an unattached system "))
	"")
    ""))

(defun staff-context (stf)
  (if stf
      (let ((pos (and *cmn-system* 
		      (staves *cmn-system*) 
		      (> (length (staves *cmn-system*)) 1)
		      (position stf (staves *cmn-system*)))))
	(format nil "~A(staff~A~A~A "
		(if pos (format nil "in staff ~D " pos) "")
		(if (staff-size stf) (format nil " (staff-size ~1,3F)" (staff-size stf)) "")
		(if (staff-lines stf) (format nil " (staff-lines ~D)" (staff-lines stf)) "")
		(if (staff-name stf) (format nil " (staff-name ~S)" (staff-name stf)) "")))
    ""))

(defun staff-data-start (stf obj owner)
  (let ((pos (and stf obj (staff-data stf)
		  (or (and owner (position owner (staff-data stf)))
		      (position obj (staff-data stf))))))
    (if (and stf (staff-data stf))
	(format nil "~{~A ~}~A"
		(loop for i from 0 below (min (or pos 5) 5) and note in (staff-data stf)  
		 if (and (not (staff-p note)) (not (bar-p note)) (not (eq obj note)) (not (eq owner note)))
		 collect (identify note))
		(if (or (and pos (> pos 8)) 
			(and (not pos) (> (length (staff-data stf)) 8)))
		    " ... " ""))
      "")))

(defun staff-data-continuation (stf obj owner sys stf0)
  (let ((pos (and stf obj (staff-data stf)
		  (or (and owner (position owner (staff-data stf)))
		      (position obj (staff-data stf))
		      (length (staff-data stf))))))
    (if (and pos (> pos 4))
	(let ((stf1 (format nil "~{~A ~}" (loop for n from (max 5 (- pos 3)) below pos
					   collect (identify (nth n (staff-data stf)))))))
	  (format nil "~A~%~VT"		;~V on p582 CLtL
		  stf1 
		  (+ 2 (length sys) (length stf) (length stf0) (length stf1))))
      (cmn-pipe-context))))

(defun owner-context (obj owner)
  (format nil "~A~A~A"
	  (if obj (identify obj))
	  (if owner " in " "")
	  (if owner (identify owner) "")))

(defun cmn-error-1 (format-string &rest args)
  (let* ((sys (system-context *cmn-system*))
	 (stf (staff-context *cmn-staff*))
	 (stf0 (staff-data-start (or *cmn-staff*
				     (and *cmn-staff-data*
					  (make-staff :data (reverse *cmn-staff-data*))))
				 *cmn-object* *cmn-owning-object*))
	 (stf1 (staff-data-continuation (or *cmn-staff* 
					    (and *cmn-staff-data*
						 (make-staff :data (reverse *cmn-staff-data*))))
					*cmn-object* *cmn-owning-object*
					sys stf0))
	 (obj (identify *cmn-object*))
	 (owner (if *cmn-owning-object* (identify *cmn-owning-object*) "")))
    (format nil "~A:~%  ~A~A~A~A~A"
	   (if args
	       (apply #'format nil format-string args)
	     format-string)
	   sys stf stf0 stf1 (owner-context obj owner))))

(defun cmn-error (format-string &rest args)
  (error (apply #'cmn-error-1 format-string args)))

(defun cmn-warn (format-string &rest args)
  (warn (apply #'cmn-error-1 format-string args)))

(defmethod notify :before (any-object &optional objects)
  (declare (ignore objects))
  (setf *cmn-object* any-object))

(defmethod display :before (any-object container score &rest rest)
  (declare (ignore score rest))
  (setf *cmn-object* any-object)
  (setf *cmn-owning-object* container))

(defmethod copy :before (any-object &optional other-object)
  (setf *cmn-object* any-object)
  (setf *cmn-owning-object* other-object))

(defmethod house :before (any-object score)
  (declare (ignore score))
  (setf *cmn-object* any-object))




;;;
;;; ----------------    included .eps files
;;;
;;; see graphics function in cmn1.lisp

(defun include-eps-defs (server)
  (let ((cmd
	 "/BEGINEPSFILE { %def
    /EPSFsave save def
    0 setgray 0 setlinecap 1 setlinewidth 0 setlinejoin 10 setmiterlimit [] 0 setdash
    newpath
    /showpage {} def
} bind def
/ENDEPSFILE { %def
    EPSFsave restore
} bind def
"))
    (if (c-output-in-progress server)
	(c-print cmd)
      (format (output server) "~A" cmd))))

(defun make-ISO-encoded-version (server font-name new-font-name)
  (let ((cmd (format nil
		     " /~A findfont
 dup length dict begin
   {1 index /FID ne {def} {pop pop} ifelse} forall
   /Encoding ISOLatin1Encoding def
   currentdict
 end
 /~A exch definefont pop
" 
		     font-name new-font-name)))
    (if (c-output-in-progress server)
	(c-print cmd)
      (format (output server) "~A" cmd))))

(defun include-eps-defs-and-whatever (server)
  (include-eps-defs server)
  (if (c-output-in-progress server) (c-print "/cmncolor 0 def") (format (output server) " /cmncolor 0 def~%"))
  (if (prologue server)
      (loop for head in (reverse (prologue server)) do
	(funcall head server))))

(defun read-4-numbers (gbb)
  (let ((gbb-as-list nil))
    (loop for i from 0 to 3 do
      (multiple-value-bind 
	  (int len)
	  (read-from-string gbb)
	(push int gbb-as-list)
	(if (< len (length gbb)) (setf gbb (subseq gbb len)))))
    (nreverse gbb-as-list)))

(defun include-eps (server file x0 y0 &optional (xscl 1.0) (yscl 1.0))
  (let ((gbb nil)
	(size (scr-size server))
	(c-out (c-output-in-progress server)))
    (if c-out
	(c-print (format nil "BEGINEPSFILE~%  ~A ~A translate~% ~1,3F ~1,3F scale" 
			 (not-rational (* x0 size))
			 (not-rational (* y0 size))
			 xscl yscl))
      (format (output server) "BEGINEPSFILE~%  ~A ~A translate~% ~1,3F ~1,3F scale~%" 
	      (not-rational (* x0 size))
	      (not-rational (* y0 size))
	      xscl yscl))
    (setf gbb (read-4-numbers (insert-eps-file server file)))
    (if c-out
	(c-print "ENDEPSFILE")
      (format (output server) "ENDEPSFILE~%"))
    (list (* x0 size)
	  (* y0 size) 
	  (+ (* x0 size) (* xscl (third gbb))) 
	  (+ (* y0 size) (* yscl (fourth gbb))))))

(defun include-eps-with-matrix-and-pattern (server file x0 y0 matrix pattern)
  (let ((gbb nil)
	(local-matrix (copy-list matrix))
	(pstart (if pattern (funcall pattern server pattern) ""))
	(c-out (c-output-in-progress server)))
    (setf (fifth local-matrix) (* x0 (scr-size server)))
    (setf (sixth local-matrix) (* y0 (scr-size server)))
    (if c-out
	(c-print (format nil "BEGINEPSFILE~% ~A [ ~{~A ~}] concat~%" pstart (map 'list #'not-rational local-matrix)))
      (format (output server) "BEGINEPSFILE~% ~A [ ~{~A ~}] concat~%" pstart (map 'list #'not-rational local-matrix)))
    (setf gbb (read-4-numbers (insert-eps-file server file)))
    (if c-out
	(c-print "ENDEPSFILE")
      (format (output server) "ENDEPSFILE~%"))
    (transform-box matrix (list (* x0 (scr-size server)) (* y0 (scr-size server))) gbb)))

(defun insert-eps-file (server file)
  ;; ideally we would leave the PostScript comments in the inserted file alone, but
  ;; that confused WriteNow which believes the first %%BoundingBox it sees and it 
  ;; quits upon encountering %%Trailer.  So, we'll strip out those portions...
  (let ((gbb nil))
    (with-open-file (s file)
      (loop do
	(let ((text (read-line s nil :at-end)))
	  (if (not (eq text :at-end))
	      (progn
		(if (and (> (length text) 15) 
			 (string= "%%BoundingBox" (subseq text 0 13)))
		    (setf gbb (subseq text 14))
		  (if (or (< (length text) 9) 
			  (not (string= "%%Trailer" (subseq text 0 9))))
		      (if (c-output-in-progress server)
			  (c-print text)
			(format (output server) "~A~%" text)))))
	    (return-from insert-eps-file gbb)))))))

(defun get-eps-bounding-box (file)
  (with-open-file (s file)
    (loop do
      (let ((text (read-line s nil :at-end)))
	(if (and (not (eq text :at-end))
		 (> (length text) 25) 
		 (string= "%%BoundingBox" (subseq text 0 13)))
	    (return-from get-eps-bounding-box (read-4-numbers (subseq text 14))))))))

(defun output-page-break (score page)
  ;; output a page break to the .ps file
  (if (c-output-in-progress score)
      (progn
	(c-print " showpage")
	(c-print (format nil "%%Page: ~D ~D" (1+ page) (1+ page)))
	(c-print (format nil " /~A findfont ~D scalefont setfont~% 0 setlinewidth" Music-Font (scr-size score))))
    (progn 
      (format (output score) " showpage~%")
      (format (output score) "%%Page: ~D ~D~%" (1+ page) (1+ page))
      (format (output score) " /~A findfont ~D scalefont setfont~% 0 setlinewidth~%" Music-Font (scr-size score)))))


(defun gray-scale (val)
  (pattern #'(lambda (score obj)
	       (declare (ignore score obj))
	       (values 
		(format nil " /cmncolor ~1,3F def  ~1,3F setgray " val val)
		" show /cmncolor 0 def 0 setgray "
		" /cmncolor 0 def 0 setgray "))))

(defun outlined (num)
  (pattern #'(lambda (score obj)
	       (declare (ignore score obj))
	       (values
		""
		(format nil " false charpath ~D setlinewidth stroke 0 setlinewidth " num)
		""))))


(defun set-gray-scale (score num)
  (let ((snum (if (integerp num) (format nil "~D" num) (format nil "~1,2F" num))))
    (if (c-output-in-progress score)
	(c-print (format nil " /cmncolor ~A def  ~A setgray~%" snum snum))
      (format (output score) " /cmncolor ~A def  ~A setgray~%" snum snum))))

(defun bind-gray-scale (score num)
  (let ((snum (if (integerp num) (format nil "~D" num) (format nil "~1,2F" num))))
    (if (c-output-in-progress score)
	(c-print (format nil " ~A setgray~%" snum))
      (format (output score) " ~A setgray~%" snum))))

(defun unbind-gray-scale (score)
  (if (c-output-in-progress score)
      (c-print (format nil " cmncolor setgray~%"))
    (format (output score) " cmncolor setgray~%")))



#|
;;; tests
(cmn (size 100) (staff (staff-lines 0) (treble (gray-scale .75)) 
     (c4 w (text "hi" (gray-scale .5)) (begin-tie (gray-scale .5)) (marcato (dy .125) (gray-scale .5) (dx .75))) 
     (c4 w end-tie) (bass (outlined 1)) g2 w))
|#